"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.configureWebSocketServices = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const json_stringify_safe_1 = __importDefault(require("json-stringify-safe"));
const semver = __importStar(require("semver"));
const core = __importStar(require("@fluidframework/server-services-core"));
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
const summarizerClientType = "summarizer";
function getRoomId(room) {
    return `${room.tenantId}/${room.documentId}`;
}
const getMessageMetadata = (documentId, tenantId) => ({
    documentId,
    tenantId,
});
const handleServerError = async (logger, errorMessage, documentId, tenantId) => {
    logger.error(errorMessage, { messageMetaData: getMessageMetadata(documentId, tenantId) });
    server_services_telemetry_1.Lumberjack.error(errorMessage, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId));
    throw new server_services_client_1.NetworkError(500, "Failed to connect client to document.");
};
const getSocketConnectThrottleId = (tenantId) => `${tenantId}_OpenSocketConn`;
const getSubmitOpThrottleId = (clientId, tenantId) => `${clientId}_${tenantId}_SubmitOp`;
const getSubmitSignalThrottleId = (clientId, tenantId) => `${clientId}_${tenantId}_SubmitSignal`;
// Sanitize the received op before sending.
function sanitizeMessage(message) {
    const sanitizedMessage = {
        clientSequenceNumber: message.clientSequenceNumber,
        contents: message.contents,
        metadata: message.metadata,
        referenceSequenceNumber: message.referenceSequenceNumber,
        traces: message.traces,
        type: message.type,
        compression: message.compression,
    };
    return sanitizedMessage;
}
const protocolVersions = ["^0.4.0", "^0.3.0", "^0.2.0", "^0.1.0"];
function selectProtocolVersion(connectVersions) {
    for (const connectVersion of connectVersions) {
        for (const protocolVersion of protocolVersions) {
            if (semver.intersects(protocolVersion, connectVersion)) {
                return protocolVersion;
            }
        }
    }
}
/**
 * Converts a relayUserAgent string into a \<key,value\> map.
 * @param relayUserAgent - user agent string in the format "prop1:val1;prop2:val2;prop3:val3"
 */
function parseRelayUserAgent(relayUserAgent) {
    if (!relayUserAgent) {
        return {};
    }
    const map = {};
    const propertyKeyValuePairs = relayUserAgent
        .split(";")
        .map((keyValuePair) => keyValuePair.split(":"));
    // TODO: would be cleaner with `Object.fromEntries()` but tsconfig needs es2019 lib
    for (const [key, value] of propertyKeyValuePairs) {
        map[key] = value;
    }
    return map;
}
/**
 * Stores client connectivity time in a Redis list.
 */
async function storeClientConnectivityTime(clientId, documentId, tenantId, connectionTimestamp, throttleAndUsageStorageManager) {
    try {
        const now = Date.now();
        const connectionTimeInMinutes = (now - connectionTimestamp) / 60000;
        const storageId = core.clientConnectivityStorageId;
        const usageData = {
            value: connectionTimeInMinutes,
            tenantId,
            documentId,
            clientId,
            startTime: connectionTimestamp,
            endTime: now,
        };
        await throttleAndUsageStorageManager.setUsageData(storageId, usageData);
    }
    catch (error) {
        server_services_telemetry_1.Lumberjack.error(`ClientConnectivity data storage failed`, {
            [server_services_telemetry_1.CommonProperties.clientId]: clientId,
            [server_services_telemetry_1.BaseTelemetryProperties.tenantId]: tenantId,
            [server_services_telemetry_1.BaseTelemetryProperties.documentId]: documentId,
        }, error);
    }
}
/**
 * @returns ThrottlingError if throttled; undefined if not throttled or no throttler provided.
 */
function checkThrottleAndUsage(throttler, throttleId, tenantId, logger, usageStorageId, usageData, incrementWeight = 1) {
    if (!throttler) {
        return;
    }
    try {
        throttler.incrementCount(throttleId, incrementWeight, usageStorageId, usageData);
    }
    catch (error) {
        if (error instanceof core.ThrottlingError) {
            return error;
        }
        else {
            logger === null || logger === void 0 ? void 0 : logger.error(`Throttle increment failed: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`, {
                messageMetaData: {
                    key: throttleId,
                    eventName: "throttling",
                },
            });
            server_services_telemetry_1.Lumberjack.error(`Throttle increment failed`, {
                [server_services_telemetry_1.CommonProperties.telemetryGroupName]: "throttling",
                [server_services_telemetry_1.BaseTelemetryProperties.tenantId]: tenantId,
            }, error);
        }
    }
}
function configureWebSocketServices(webSocketServer, orderManager, tenantManager, storage, clientManager, metricLogger, logger, maxNumberOfClientsPerDocument = 1000000, numberOfMessagesPerTrace = 100, maxTokenLifetimeSec = 60 * 60, isTokenExpiryEnabled = false, isClientConnectivityCountingEnabled = false, isSignalUsageCountingEnabled = false, cache, connectThrottlerPerTenant, connectThrottlerPerCluster, submitOpThrottler, submitSignalThrottler, throttleAndUsageStorageManager, verifyMaxMessageSize, socketTracker, revokedTokenChecker) {
    webSocketServer.on("connection", (socket) => {
        // Map from client IDs on this connection to the object ID and user info.
        const connectionsMap = new Map();
        // Map from client IDs to room.
        const roomMap = new Map();
        // Map from client Ids to scope.
        const scopeMap = new Map();
        // Map from client Ids to connection time.
        const connectionTimeMap = new Map();
        let connectDocumentComplete = false;
        let connectDocumentP;
        const clientIdConnectionsDisconnected = new Set();
        const clientIdClientsDisconnected = new Set();
        let disconnectDocumentP;
        // Timer to check token expiry for this socket connection
        let expirationTimer;
        const connectionCountLogger = new utils_1.ConnectionCountLogger(process.env.NODE_NAME, cache);
        const hasWriteAccess = (scopes) => (0, server_services_client_1.canWrite)(scopes) || (0, server_services_client_1.canSummarize)(scopes);
        function isWriter(scopes, mode) {
            return hasWriteAccess(scopes) ? mode === "write" : false;
        }
        function clearExpirationTimer() {
            if (expirationTimer !== undefined) {
                clearTimeout(expirationTimer);
                expirationTimer = undefined;
            }
        }
        function setExpirationTimer(mSecUntilExpiration) {
            clearExpirationTimer();
            expirationTimer = setTimeout(() => {
                socket.disconnect(true);
            }, mSecUntilExpiration);
        }
        async function connectDocument(message) {
            var _a;
            const startTime = Date.now();
            const throttleErrorPerCluster = checkThrottleAndUsage(connectThrottlerPerCluster, getSocketConnectThrottleId("connectDoc"), message.tenantId, logger);
            if (throttleErrorPerCluster) {
                // eslint-disable-next-line @typescript-eslint/no-throw-literal
                throw throttleErrorPerCluster;
            }
            const throttleErrorPerTenant = checkThrottleAndUsage(connectThrottlerPerTenant, getSocketConnectThrottleId(message.tenantId), message.tenantId, logger);
            if (throttleErrorPerTenant) {
                // eslint-disable-next-line @typescript-eslint/no-throw-literal
                throw throttleErrorPerTenant;
            }
            if (!message.token) {
                throw new server_services_client_1.NetworkError(403, "Must provide an authorization token");
            }
            // Validate token signature and claims
            const token = message.token;
            const claims = (0, server_services_client_1.validateTokenClaims)(token, message.id, message.tenantId);
            const clientId = (0, utils_1.generateClientId)();
            const lumberjackProperties = Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(claims.documentId, claims.tenantId)), { [server_services_telemetry_1.CommonProperties.clientId]: clientId });
            const connectDocumentGetClientsMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentGetClients, lumberjackProperties);
            const clients = await clientManager
                .getClients(claims.tenantId, claims.documentId)
                .then((response) => {
                connectDocumentGetClientsMetric.success("Successfully got clients from client manager");
                return response;
            })
                .catch(async (err) => {
                const errMsg = `Failed to get clients. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
                connectDocumentGetClientsMetric.error("Failed to get clients during connectDocument", err);
                return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
            });
            if (clients.length > maxNumberOfClientsPerDocument) {
                throw new server_services_client_1.NetworkError(429, "Too Many Clients Connected to Document", true /* canRetry */, false /* isFatal */, 5 * 60 * 1000 /* retryAfterMs (5 min) */);
            }
            try {
                // Check if token is revoked
                if (revokedTokenChecker && claims.jti) {
                    const isTokenRevoked = await revokedTokenChecker.isTokenRevoked(claims.tenantId, claims.documentId, claims.jti);
                    if (isTokenRevoked) {
                        throw new core.TokenRevokedError(403, "Permission denied. Token has been revoked", false /* canRetry */, true /* isFatal */);
                    }
                }
                await tenantManager.verifyToken(claims.tenantId, token);
            }
            catch (error) {
                if ((0, server_services_client_1.isNetworkError)(error)) {
                    throw error;
                }
                // We don't understand the error, so it is likely an internal service error.
                const errMsg = `Could not verify connect document token. Error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`;
                return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
            }
            const room = {
                tenantId: claims.tenantId,
                documentId: claims.documentId,
            };
            try {
                // Subscribe to channels.
                await Promise.all([
                    socket.join(getRoomId(room)),
                    socket.join(`client#${clientId}`),
                ]);
            }
            catch (err) {
                const errMsg = `Could not subscribe to channels. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
                return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
            }
            const connectedTimestamp = Date.now();
            // Todo: should all the client details come from the claims???
            // we are still trusting the users permissions and type here.
            const messageClient = message.client ? message.client : {};
            const isSummarizer = ((_a = messageClient.details) === null || _a === void 0 ? void 0 : _a.type) === summarizerClientType;
            messageClient.user = claims.user;
            messageClient.scopes = claims.scopes;
            // 1. Do not give SummaryWrite scope to clients that are not summarizers.
            // 2. Store connection timestamp for all clients but the summarizer.
            // Connection timestamp is used (inside socket disconnect event) to
            // calculate the client connection time (i.e. for billing).
            if (!isSummarizer) {
                messageClient.scopes = claims.scopes.filter((scope) => scope !== protocol_definitions_1.ScopeType.SummaryWrite);
                connectionTimeMap.set(clientId, connectedTimestamp);
            }
            // back-compat: remove cast to any once new definition of IClient comes through.
            messageClient.timestamp = connectedTimestamp;
            // Cache the scopes.
            scopeMap.set(clientId, messageClient.scopes);
            // Join the room to receive signals.
            roomMap.set(clientId, room);
            // increment connection count after the client is added to the room.
            // excluding summarizer for total client count.
            if (!isSummarizer) {
                connectionCountLogger.incrementConnectionCount();
            }
            // Iterate over the version ranges provided by the client and select the best one that works
            const connectVersions = message.versions ? message.versions : ["^0.1.0"];
            const version = selectProtocolVersion(connectVersions);
            if (!version) {
                throw new server_services_client_1.NetworkError(400, `Unsupported client protocol. Server: ${protocolVersions}. Client: ${JSON.stringify(connectVersions)}`);
            }
            const connectDocumentAddClientMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentAddClient, lumberjackProperties);
            try {
                await clientManager.addClient(claims.tenantId, claims.documentId, clientId, messageClient);
                connectDocumentAddClientMetric.success("Successfully added client");
            }
            catch (err) {
                const errMsg = `Could not add client. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
                connectDocumentAddClientMetric.error("Error adding client during connectDocument", err);
                return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
            }
            if (isTokenExpiryEnabled) {
                const lifeTimeMSec = (0, server_services_client_1.validateTokenClaimsExpiration)(claims, maxTokenLifetimeSec);
                setExpirationTimer(lifeTimeMSec);
            }
            let connectedMessage;
            if (isWriter(messageClient.scopes, message.mode)) {
                const connectDocumentOrdererConnectionMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocumentOrdererConnection, lumberjackProperties);
                const orderer = await orderManager
                    .getOrderer(claims.tenantId, claims.documentId)
                    .catch(async (err) => {
                    const errMsg = `Failed to get orderer manager. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
                    connectDocumentOrdererConnectionMetric.error("Failed to get orderer manager", err);
                    return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
                });
                const connection = await orderer
                    .connect(socket, clientId, messageClient)
                    .catch(async (err) => {
                    const errMsg = `Failed to connect to orderer. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
                    connectDocumentOrdererConnectionMetric.error("Failed to connect to orderer", err);
                    return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
                });
                // Eventually we will send disconnect reason as headers to client.
                connection.once("error", (error) => {
                    const messageMetaData = getMessageMetadata(connection.documentId, connection.tenantId);
                    logger.error(`Disconnecting socket on connection error: ${(0, json_stringify_safe_1.default)(error, undefined, 2)}`, { messageMetaData });
                    server_services_telemetry_1.Lumberjack.error(`Disconnecting socket on connection error`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId), error);
                    clearExpirationTimer();
                    socket.disconnect(true);
                });
                let clientJoinMessageServerMetadata;
                if (core.DefaultServiceConfiguration.enableTraces &&
                    sampleMessages(numberOfMessagesPerTrace)) {
                    clientJoinMessageServerMetadata = {
                        connectDocumentStartTime: startTime,
                    };
                }
                connection.connect(clientJoinMessageServerMetadata).catch(async (err) => {
                    const errMsg = `Failed to connect to the orderer connection. Error: ${(0, json_stringify_safe_1.default)(err, undefined, 2)}`;
                    connectDocumentOrdererConnectionMetric.error("Failed to establish orderer connection", err);
                    return handleServerError(logger, errMsg, claims.documentId, claims.tenantId);
                });
                connectionsMap.set(clientId, connection);
                if (connectionsMap.size > 1) {
                    server_services_telemetry_1.Lumberjack.info(`Same socket is having multiple connections, connection number=${connectionsMap.size}`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId));
                }
                connectDocumentOrdererConnectionMetric.success("Successfully established orderer connection");
                connectedMessage = {
                    claims,
                    clientId,
                    existing: true,
                    maxMessageSize: connection.maxMessageSize,
                    mode: "write",
                    serviceConfiguration: {
                        blockSize: connection.serviceConfiguration.blockSize,
                        maxMessageSize: connection.serviceConfiguration.maxMessageSize,
                    },
                    initialClients: clients,
                    initialMessages: [],
                    initialSignals: [],
                    supportedVersions: protocolVersions,
                    version,
                };
            }
            else {
                connectedMessage = {
                    claims,
                    clientId,
                    existing: true,
                    maxMessageSize: 1024,
                    mode: "read",
                    serviceConfiguration: {
                        blockSize: core.DefaultServiceConfiguration.blockSize,
                        maxMessageSize: core.DefaultServiceConfiguration.maxMessageSize,
                    },
                    initialClients: clients,
                    initialMessages: [],
                    initialSignals: [],
                    supportedVersions: protocolVersions,
                    version,
                };
            }
            // back-compat: remove cast to any once new definition of IConnected comes through.
            connectedMessage.timestamp = connectedTimestamp;
            // Track socket and tokens for this connection
            if (socketTracker && claims.jti) {
                socketTracker.addSocketForToken(core.createCompositeTokenId(message.tenantId, message.id, claims.jti), socket);
            }
            return {
                connection: connectedMessage,
                connectVersions,
                details: messageClient,
            };
        }
        async function disconnectDocument() {
            // Clear token expiration timer on disconnection
            clearExpirationTimer();
            const removeAndStoreP = [];
            // Send notification messages for all client IDs in the connection map
            for (const [clientId, connection] of connectionsMap) {
                if (clientIdConnectionsDisconnected.has(clientId)) {
                    // We already removed this clientId once. Skip it.
                    continue;
                }
                const messageMetaData = getMessageMetadata(connection.documentId, connection.tenantId);
                logger.info(`Disconnect of ${clientId}`, { messageMetaData });
                server_services_telemetry_1.Lumberjack.info(`Disconnect of ${clientId}`, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId));
                connection
                    .disconnect()
                    .then(() => {
                    // Keep track of disconnected clientIds so that we don't repeat the disconnect signal
                    // for the same clientId if retrying when connectDocument completes after disconnectDocument.
                    clientIdConnectionsDisconnected.add(clientId);
                })
                    .catch((error) => {
                    const errorMsg = `Failed to disconnect client ${clientId} from orderer connection.`;
                    server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId), error);
                });
                if (isClientConnectivityCountingEnabled && throttleAndUsageStorageManager) {
                    const connectionTimestamp = connectionTimeMap.get(clientId);
                    if (connectionTimestamp) {
                        removeAndStoreP.push(storeClientConnectivityTime(clientId, connection.documentId, connection.tenantId, connectionTimestamp, throttleAndUsageStorageManager));
                    }
                }
            }
            // Send notification messages for all client IDs in the room map
            for (const [clientId, room] of roomMap) {
                if (clientIdClientsDisconnected.has(clientId)) {
                    // We already removed this clientId once. Skip it.
                    continue;
                }
                const messageMetaData = getMessageMetadata(room.documentId, room.tenantId);
                // excluding summarizer for total client count.
                if (connectionTimeMap.has(clientId)) {
                    connectionCountLogger.decrementConnectionCount();
                }
                logger.info(`Disconnect of ${clientId} from room`, { messageMetaData });
                server_services_telemetry_1.Lumberjack.info(`Disconnect of ${clientId} from room`, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId));
                removeAndStoreP.push(clientManager
                    .removeClient(room.tenantId, room.documentId, clientId)
                    .then(() => {
                    // Keep track of disconnected clientIds so that we don't repeat the disconnect signal
                    // for the same clientId if retrying when connectDocument completes after disconnectDocument.
                    clientIdClientsDisconnected.add(clientId);
                }));
                socket
                    .emitToRoom(getRoomId(room), "signal", (0, utils_1.createRoomLeaveMessage)(clientId))
                    .catch((error) => {
                    const errorMsg = `Failed to emit signal to room ${clientId}, ${getRoomId(room)}.`;
                    server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(room.documentId, room.tenantId), error);
                });
            }
            // Clear socket tracker upon disconnection
            if (socketTracker) {
                socketTracker.removeSocket(socket.id);
            }
            await Promise.all(removeAndStoreP);
        }
        // Note connect is a reserved socket.io word so we use connect_document to represent the connect request
        // eslint-disable-next-line @typescript-eslint/no-misused-promises
        socket.on("connect_document", async (connectionMessage) => {
            const userAgentInfo = parseRelayUserAgent(connectionMessage.relayUserAgent);
            const driverVersion = userAgentInfo.driverVersion;
            const connectMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ConnectDocument);
            const baseLumberjackProperties = (0, server_services_telemetry_1.getLumberBaseProperties)(connectionMessage.id, connectionMessage.tenantId);
            connectMetric.setProperties(Object.assign(Object.assign({}, baseLumberjackProperties), { [server_services_telemetry_1.CommonProperties.clientDriverVersion]: driverVersion, [server_services_telemetry_1.CommonProperties.connectionCount]: connectionsMap.size, [server_services_telemetry_1.CommonProperties.connectionClients]: JSON.stringify(Array.from(connectionsMap.keys())), [server_services_telemetry_1.CommonProperties.roomClients]: JSON.stringify(Array.from(roomMap.keys())) }));
            connectDocumentP = connectDocument(connectionMessage)
                .then((message) => {
                var _a;
                socket.emit("connect_document_success", message.connection);
                const room = roomMap.get(message.connection.clientId);
                if (room) {
                    socket.emitToRoom(getRoomId(room), "signal", (0, utils_1.createRoomJoinMessage)(message.connection.clientId, message.details));
                }
                connectMetric.setProperties({
                    [server_services_telemetry_1.CommonProperties.clientId]: message.connection.clientId,
                    [server_services_telemetry_1.CommonProperties.clientCount]: message.connection.initialClients.length + 1,
                    [server_services_telemetry_1.CommonProperties.clientType]: (_a = message.details.details) === null || _a === void 0 ? void 0 : _a.type,
                });
                connectMetric.success(`Connect document successful`);
            })
                .catch((error) => {
                socket.emit("connect_document_error", error);
                clearExpirationTimer();
                if ((error === null || error === void 0 ? void 0 : error.code) !== undefined) {
                    connectMetric.setProperty(server_services_telemetry_1.CommonProperties.errorCode, error.code);
                }
                connectMetric.error(`Connect document failed`, error);
            })
                .finally(() => {
                connectDocumentComplete = true;
                if (disconnectDocumentP) {
                    server_services_telemetry_1.Lumberjack.warning(`ConnectDocument completed after disconnect was handled.`);
                    // We have already received disconnect for this connection.
                    disconnectDocumentP
                        .catch((error) => {
                        server_services_telemetry_1.Lumberjack.error("Error encountered in disconnectDocumentP", Object.assign({}, baseLumberjackProperties), error);
                    })
                        .finally(() => {
                        // We might need to re-run disconnect handler after previous disconnect handler completes.
                        // DisconnectDocument internally handles the cases where we have already run disconnect for
                        // roomsMap and connectionsMap so that we don't duplicate disconnect efforts.
                        // The primary need for this retry is when we receive "disconnect" in the narrow window after
                        // "connect_document" but before "connectDocumentP" is defined.
                        const alreadyDisconnectedAllConnections = connectionsMap.size === clientIdConnectionsDisconnected.size;
                        const alreadyDisconnectedAllClients = roomMap.size === clientIdClientsDisconnected.size;
                        if (alreadyDisconnectedAllConnections &&
                            alreadyDisconnectedAllClients) {
                            // Don't retry disconnect if all connections and clients are already handled.
                            return;
                        }
                        const disconnectRetryMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.DisconnectDocumentRetry);
                        disconnectRetryMetric.setProperties(Object.assign(Object.assign({}, baseLumberjackProperties), { [server_services_telemetry_1.CommonProperties.connectionCount]: connectionsMap.size, [server_services_telemetry_1.CommonProperties.connectionClients]: JSON.stringify(Array.from(connectionsMap.keys())), [server_services_telemetry_1.CommonProperties.roomClients]: JSON.stringify(Array.from(roomMap.keys())) }));
                        disconnectDocument()
                            .then(() => {
                            disconnectRetryMetric.success(`Successfully retried disconnect.`);
                        })
                            .catch((error) => {
                            disconnectRetryMetric.error(`Disconnect retry failed.`, error);
                        });
                    });
                }
            });
        });
        // Message sent when a new operation is submitted to the router
        socket.on("submitOp", (clientId, messageBatches) => {
            // Verify the user has an orderer connection.
            const connection = connectionsMap.get(clientId);
            if (!connection) {
                let nackMessage;
                const clientScope = scopeMap.get(clientId);
                if (clientScope && hasWriteAccess(clientScope)) {
                    nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Readonly client");
                }
                else if (roomMap.has(clientId)) {
                    nackMessage = (0, utils_1.createNackMessage)(403, protocol_definitions_1.NackErrorType.InvalidScopeError, "Invalid scope");
                }
                else {
                    nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Nonexistent client");
                }
                socket.emit("nack", "", [nackMessage]);
            }
            else {
                let messageCount = 0;
                for (const messageBatch of messageBatches) {
                    // Count all messages in each batch for accurate throttling calculation.
                    // Note: This is happening before message size checking. We won't process
                    // messages that are too large, so it is inaccurate to increment throttle
                    // counts for unprocessed messages.
                    messageCount += Array.isArray(messageBatch) ? messageBatch.length : 1;
                }
                const throttleError = checkThrottleAndUsage(submitOpThrottler, getSubmitOpThrottleId(clientId, connection.tenantId), connection.tenantId, logger, undefined, undefined, messageCount /* incrementWeight */);
                if (throttleError) {
                    const nackMessage = (0, utils_1.createNackMessage)(throttleError.code, protocol_definitions_1.NackErrorType.ThrottlingError, throttleError.message, throttleError.retryAfter);
                    socket.emit("nack", "", [nackMessage]);
                    return;
                }
                const lumberjackProperties = Object.assign({ [server_services_telemetry_1.CommonProperties.clientId]: clientId }, (0, server_services_telemetry_1.getLumberBaseProperties)(connection.documentId, connection.tenantId));
                const handleMessageBatchProcessingError = (error) => {
                    if ((0, server_services_client_1.isNetworkError)(error)) {
                        if (error.code === 413) {
                            server_services_telemetry_1.Lumberjack.info("Rejected too large operation(s)", lumberjackProperties);
                            socket.emit("nack", "", [
                                (0, utils_1.createNackMessage)(error.code, protocol_definitions_1.NackErrorType.BadRequestError, error.message),
                            ]);
                            return;
                        }
                    }
                    server_services_telemetry_1.Lumberjack.error("Error processing submitted op(s)", lumberjackProperties, error);
                };
                messageBatches.forEach((messageBatch) => {
                    const messages = Array.isArray(messageBatch)
                        ? messageBatch
                        : [messageBatch];
                    try {
                        const sanitized = messages.map((message) => {
                            if (verifyMaxMessageSize === true) {
                                // Local tests show `JSON.stringify` to be fast
                                // - <1ms for JSONs <100kb
                                // - ~2ms for JSONs ~256kb
                                // - ~6ms for JSONs ~1mb
                                // - ~38ms for JSONs ~10mb
                                // - ~344ms for JSONs ~100mb
                                // maxMessageSize is currently 16kb, so this check should be <1ms
                                const messageSize = JSON.stringify(message.contents).length;
                                const maxMessageSize = connection.serviceConfiguration.maxMessageSize;
                                if (messageSize > maxMessageSize) {
                                    // Exit early from processing message batch
                                    throw new server_services_client_1.NetworkError(413, "Op size too large");
                                }
                            }
                            const sanitizedMessage = sanitizeMessage(message);
                            const sanitizedMessageWithTrace = addAlfredTrace(sanitizedMessage, numberOfMessagesPerTrace, connection.clientId, connection.tenantId, connection.documentId);
                            return sanitizedMessageWithTrace;
                        });
                        if (sanitized.length > 0) {
                            // Cannot await this order call without delaying other message batches in this submitOp.
                            connection
                                .order(sanitized)
                                .catch(handleMessageBatchProcessingError);
                        }
                    }
                    catch (e) {
                        handleMessageBatchProcessingError(e);
                    }
                });
            }
        });
        // Message sent when a new signal is submitted to the router
        socket.on("submitSignal", (clientId, contentBatches) => {
            // Verify the user has subscription to the room.
            const room = roomMap.get(clientId);
            if (!room) {
                const nackMessage = (0, utils_1.createNackMessage)(400, protocol_definitions_1.NackErrorType.BadRequestError, "Nonexistent client");
                socket.emit("nack", "", [nackMessage]);
            }
            else {
                let messageCount = 0;
                for (const contentBatch of contentBatches) {
                    // Count all messages in each batch for accurate throttling calculation.
                    messageCount += Array.isArray(contentBatch) ? contentBatch.length : 1;
                }
                const signalUsageData = {
                    value: 0,
                    tenantId: room.tenantId,
                    documentId: room.documentId,
                    clientId,
                };
                const throttleError = checkThrottleAndUsage(submitSignalThrottler, getSubmitSignalThrottleId(clientId, room.tenantId), room.tenantId, logger, isSignalUsageCountingEnabled ? core.signalUsageStorageId : undefined, isSignalUsageCountingEnabled ? signalUsageData : undefined, messageCount /* incrementWeight */);
                if (throttleError) {
                    const nackMessage = (0, utils_1.createNackMessage)(throttleError.code, protocol_definitions_1.NackErrorType.ThrottlingError, throttleError.message, throttleError.retryAfter);
                    socket.emit("nack", "", [nackMessage]);
                    return;
                }
                contentBatches.forEach((contentBatch) => {
                    const contents = Array.isArray(contentBatch)
                        ? contentBatch
                        : [contentBatch];
                    for (const content of contents) {
                        const signalMessage = {
                            clientId,
                            content,
                        };
                        socket.emitToRoom(getRoomId(room), "signal", signalMessage);
                    }
                });
            }
        });
        // eslint-disable-next-line @typescript-eslint/no-misused-promises
        socket.on("disconnect", async () => {
            if (!connectDocumentComplete && connectDocumentP) {
                server_services_telemetry_1.Lumberjack.warning(`Socket connection disconnected before ConnectDocument completed.`);
                // Wait for document connection to finish before disconnecting.
                // If disconnect fires before roomMap or connectionsMap are updated, we can be left with
                // hanging connections and clients.
                await connectDocumentP.catch(() => { });
            }
            const disconnectMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.DisconnectDocument);
            disconnectMetric.setProperties({
                [server_services_telemetry_1.CommonProperties.connectionCount]: connectionsMap.size,
                [server_services_telemetry_1.CommonProperties.connectionClients]: JSON.stringify(Array.from(connectionsMap.keys())),
                [server_services_telemetry_1.CommonProperties.roomClients]: JSON.stringify(Array.from(roomMap.keys())),
            });
            if (roomMap.size >= 1) {
                const rooms = Array.from(roomMap.values());
                const documentId = rooms[0].documentId;
                const tenantId = rooms[0].tenantId;
                disconnectMetric.setProperties(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(documentId, tenantId)));
            }
            try {
                disconnectDocumentP = disconnectDocument();
                await disconnectDocumentP;
                disconnectMetric.success(`Successfully disconnected.`);
            }
            catch (error) {
                disconnectMetric.error(`Disconnect failed.`, error);
            }
        });
    });
}
exports.configureWebSocketServices = configureWebSocketServices;
function addAlfredTrace(message, numberOfMessagesPerTrace, clientId, tenantId, documentId) {
    if (message &&
        core.DefaultServiceConfiguration.enableTraces &&
        sampleMessages(numberOfMessagesPerTrace)) {
        if (message.traces === undefined) {
            message.traces = [];
        }
        message.traces.push({
            action: "start",
            service: "alfred",
            timestamp: Date.now(),
        });
        const lumberjackProperties = {
            [server_services_telemetry_1.BaseTelemetryProperties.tenantId]: tenantId,
            [server_services_telemetry_1.BaseTelemetryProperties.documentId]: documentId,
            clientId,
            clientSequenceNumber: message.clientSequenceNumber,
            traces: message.traces,
            opType: message.type,
        };
        server_services_telemetry_1.Lumberjack.info(`Message received by Alfred.`, lumberjackProperties);
    }
    return message;
}
function sampleMessages(numberOfMessagesPerTrace) {
    return (0, server_services_client_1.getRandomInt)(numberOfMessagesPerTrace) === 0;
}
//# sourceMappingURL=index.js.map