"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.CopierLambda = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
class CopierLambda {
    constructor(rawOpCollection, context) {
        this.rawOpCollection = rawOpCollection;
        this.context = context;
        // Below, one job corresponds to the task of sending one batch to Mongo:
        this.pendingJobs = new Map();
        this.currentJobs = new Map();
    }
    handler(message) {
        // Extract batch of raw ops from Kafka message:
        const boxcar = (0, server_services_core_1.extractBoxcar)(message);
        const batch = boxcar.contents;
        const topic = `${boxcar.tenantId}/${boxcar.documentId}`;
        // Extract boxcar contents and group the ops into the message batch:
        const submittedBatch = {
            index: message.offset,
            documentId: boxcar.documentId,
            tenantId: boxcar.tenantId,
            contents: batch.map((m) => m),
        };
        // Write the batch directly to Mongo:
        let pendingJobs = this.pendingJobs.get(topic);
        if (!pendingJobs) {
            pendingJobs = [];
            this.pendingJobs.set(topic, pendingJobs);
        }
        pendingJobs.push(submittedBatch);
        // Update current offset (will be tied to this batch):
        this.pendingOffset = message;
        this.sendPending();
        return undefined;
    }
    close() {
        this.pendingJobs.clear();
        this.currentJobs.clear();
        return;
    }
    sendPending() {
        // If there is work currently being sent or we have no pending work return early
        if (this.currentJobs.size > 0 || this.pendingJobs.size === 0) {
            return;
        }
        // Swap current and pending
        const temp = this.currentJobs;
        this.currentJobs = this.pendingJobs;
        this.pendingJobs = temp;
        const batchOffset = this.pendingOffset;
        const allProcessed = [];
        // Process all current jobs on all current topics:
        for (const [, batch] of this.currentJobs) {
            const processP = this.processMongoCore(batch);
            allProcessed.push(processP);
        }
        Promise.all(allProcessed)
            .then(() => {
            this.currentJobs.clear();
            this.context.checkpoint(batchOffset);
            this.sendPending();
        })
            .catch((error) => {
            this.context.error(error, { restart: true });
        });
    }
    async processMongoCore(kafkaBatches) {
        await this.rawOpCollection.insertMany(kafkaBatches, false).catch((error) => {
            // Duplicate key errors are ignored since a replay may cause us to insert twice into Mongo.
            // All other errors result in a rejected promise.
            if (error.code !== 11000) {
                // Needs to be a full rejection here
                throw error;
            }
        });
    }
}
exports.CopierLambda = CopierLambda;
//# sourceMappingURL=lambda.js.map