"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.CheckpointContext = void 0;
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const utils_1 = require("../utils");
class CheckpointContext {
    constructor(tenantId, id, checkpointManager, context, checkpointService) {
        this.tenantId = tenantId;
        this.id = id;
        this.checkpointManager = checkpointManager;
        this.context = context;
        this.checkpointService = checkpointService;
        this.closed = false;
    }
    /**
     * Checkpoints to the database & kafka
     * Note: This is an async method, but you should not await this
     */
    async checkpoint(checkpoint, restartOnCheckpointFailure, globalCheckpointOnly) {
        var _a, _b;
        // Exit early if already closed
        if (this.closed) {
            return;
        }
        // Check if a checkpoint is in progress - if so store the pending checkpoint
        if (this.pendingUpdateP) {
            this.pendingCheckpoint = checkpoint;
            return;
        }
        // Database checkpoint
        try {
            this.pendingUpdateP = this.checkpointCore(checkpoint, globalCheckpointOnly);
            await this.pendingUpdateP;
        }
        catch (ex) {
            // TODO flag context as error / use this.context.error() instead?
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`Error writing checkpoint to the database: ${JSON.stringify(ex)}`, {
                messageMetaData: {
                    documentId: this.id,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(`Error writing checkpoint to the database`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.id, this.tenantId), ex);
            return;
        }
        // Kafka checkpoint
        try {
            // depending on the sequence of events, it might try to checkpoint the same offset a second time
            // detect and prevent that case here
            const kafkaCheckpointMessage = checkpoint.kafkaCheckpointMessage;
            if (kafkaCheckpointMessage &&
                (this.lastKafkaCheckpointOffset === undefined ||
                    kafkaCheckpointMessage.offset > this.lastKafkaCheckpointOffset)) {
                this.lastKafkaCheckpointOffset = kafkaCheckpointMessage.offset;
                this.context.checkpoint(kafkaCheckpointMessage, restartOnCheckpointFailure);
            }
        }
        catch (ex) {
            // TODO flag context as error / use this.context.error() instead?
            (_b = this.context.log) === null || _b === void 0 ? void 0 : _b.error(`Error writing checkpoint to kafka: ${JSON.stringify(ex)}`, {
                messageMetaData: {
                    documentId: this.id,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(`Error writing checkpoint to the kafka`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.id, this.tenantId), ex);
        }
        this.pendingUpdateP = undefined;
        // Trigger another round if there is a pending update
        if (this.pendingCheckpoint) {
            const pendingCheckpoint = this.pendingCheckpoint;
            this.pendingCheckpoint = undefined;
            void this.checkpoint(pendingCheckpoint);
        }
    }
    close() {
        this.closed = true;
    }
    checkpointCore(checkpoint, globalCheckpointOnly = false) {
        // Exit early if already closed
        if (this.closed) {
            return;
        }
        let updateP;
        const localCheckpointEnabled = this.checkpointService.localCheckpointEnabled;
        // determine if checkpoint is local
        const isLocal = globalCheckpointOnly === true
            ? false
            : localCheckpointEnabled && checkpoint.reason !== utils_1.CheckpointReason.NoClients;
        if (checkpoint.clear) {
            updateP = this.checkpointManager.deleteCheckpoint(checkpoint, isLocal);
        }
        else {
            // clone the checkpoint
            const deliCheckpoint = Object.assign({}, checkpoint.deliState);
            updateP = this.checkpointManager.writeCheckpoint(deliCheckpoint, isLocal, checkpoint.reason);
        }
        // Retry the checkpoint on error
        // eslint-disable-next-line @typescript-eslint/promise-function-async
        return updateP.catch((error) => {
            var _a;
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`Error writing checkpoint to MongoDB: ${JSON.stringify(error)}`, {
                messageMetaData: {
                    documentId: this.id,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(`Error writing checkpoint to MongoDB`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.id, this.tenantId), error);
            return new Promise((resolve, reject) => {
                resolve(this.checkpointCore(checkpoint));
            });
        });
    }
}
exports.CheckpointContext = CheckpointContext;
//# sourceMappingURL=checkpointContext.js.map