"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.ClientSequenceNumberManager = void 0;
const common_utils_1 = require("@fluidframework/common-utils");
const SequenceNumberComparer = {
    compare: (a, b) => a.referenceSequenceNumber - b.referenceSequenceNumber,
    min: {
        canEvict: true,
        clientId: undefined,
        clientSequenceNumber: 0,
        lastUpdate: -1,
        nack: false,
        referenceSequenceNumber: -1,
        scopes: [],
    },
};
class ClientSequenceNumberManager {
    constructor() {
        this.clientNodeMap = new Map();
        this.clientSeqNumbers = new common_utils_1.Heap(SequenceNumberComparer);
    }
    has(clientId) {
        return this.clientNodeMap.has(clientId);
    }
    get(clientId) {
        const node = this.clientNodeMap.get(clientId);
        if (node === undefined) {
            return undefined;
        }
        return node.value;
    }
    count() {
        return this.clientNodeMap.size;
    }
    peek() {
        const node = this.clientSeqNumbers.peek();
        if (node === undefined) {
            return undefined;
        }
        return node.value;
    }
    cloneValues() {
        const clients = [];
        for (const [, value] of this.clientNodeMap) {
            const source = value.value;
            clients.push(Object.assign({}, source));
        }
        return clients;
    }
    /**
     * Begins tracking or updates an already tracked client.
     * @param clientId - The client identifier
     * @param clientSequenceNumber - The sequence number generated by client
     * @param referenceSequenceNumber - The sequence number the client is at
     * @param timestamp - The time of the operation
     * @param canEvict - Flag indicating whether or not we can evict the client (branch clients cannot be evicted)
     * @param scopes - scope of the client
     * @param nack - Flag indicating whether we have nacked this client
     * @param serverMetadata - Optional server provided metadata to associate with the client
     * Returns false if the same client has been added earlier.
     */
    upsertClient(clientId, clientSequenceNumber, referenceSequenceNumber, timestamp, canEvict, scopes = [], nack = false, serverMetadata = undefined) {
        const client = this.clientNodeMap.get(clientId);
        if (client) {
            client.value.referenceSequenceNumber = referenceSequenceNumber;
            client.value.clientSequenceNumber = clientSequenceNumber;
            client.value.lastUpdate = timestamp;
            client.value.nack = nack;
            if (serverMetadata) {
                // update serverMetadata if it's provided
                client.value.serverMetadata = serverMetadata;
            }
            this.clientSeqNumbers.update(client);
            return false;
        }
        // Add the client ID to our map since this is the first time we've seen it
        const newNode = this.clientSeqNumbers.add({
            canEvict,
            clientId,
            clientSequenceNumber,
            lastUpdate: timestamp,
            nack,
            referenceSequenceNumber,
            scopes,
            serverMetadata,
        });
        this.clientNodeMap.set(clientId, newNode);
        return true;
    }
    /**
     * Removes the provided client from the list of tracked clients.
     * Returns false if the client has been removed earlier.
     */
    removeClient(clientId) {
        const details = this.clientNodeMap.get(clientId);
        if (!details) {
            return false;
        }
        // Remove the client from the list of nodes
        this.clientSeqNumbers.remove(details);
        this.clientNodeMap.delete(clientId);
        return true;
    }
    /**
     * Retrieves the minimum sequence number.
     */
    getMinimumSequenceNumber() {
        if (this.clientSeqNumbers.count() > 0) {
            const client = this.clientSeqNumbers.peek();
            return client.value.referenceSequenceNumber;
        }
        else {
            return -1;
        }
    }
}
exports.ClientSequenceNumberManager = ClientSequenceNumberManager;
//# sourceMappingURL=clientSeqManager.js.map