"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeliLambda = exports.OpEventType = void 0;
const protocol_definitions_1 = require("@fluidframework/protocol-definitions");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const common_utils_1 = require("@fluidframework/common-utils");
const utils_1 = require("../utils");
const checkpointContext_1 = require("./checkpointContext");
const clientSeqManager_1 = require("./clientSeqManager");
var IncomingMessageOrder;
(function (IncomingMessageOrder) {
    IncomingMessageOrder[IncomingMessageOrder["Duplicate"] = 0] = "Duplicate";
    IncomingMessageOrder[IncomingMessageOrder["Gap"] = 1] = "Gap";
    IncomingMessageOrder[IncomingMessageOrder["ConsecutiveOrSystem"] = 2] = "ConsecutiveOrSystem";
})(IncomingMessageOrder || (IncomingMessageOrder = {}));
var SendType;
(function (SendType) {
    SendType[SendType["Immediate"] = 0] = "Immediate";
    SendType[SendType["Later"] = 1] = "Later";
    SendType[SendType["Never"] = 2] = "Never";
})(SendType || (SendType = {}));
var InstructionType;
(function (InstructionType) {
    InstructionType[InstructionType["ClearCache"] = 0] = "ClearCache";
    InstructionType[InstructionType["NoOp"] = 1] = "NoOp";
})(InstructionType || (InstructionType = {}));
var TicketType;
(function (TicketType) {
    TicketType[TicketType["Sequenced"] = 0] = "Sequenced";
    TicketType[TicketType["Nack"] = 1] = "Nack";
    TicketType[TicketType["Signal"] = 2] = "Signal";
})(TicketType || (TicketType = {}));
var OpEventType;
(function (OpEventType) {
    /**
     * There have been no sequenced ops for X milliseconds since the last message.
     */
    OpEventType[OpEventType["Idle"] = 0] = "Idle";
    /**
     * More than X amount of ops have been ticketed since the emit.
     */
    OpEventType[OpEventType["MaxOps"] = 1] = "MaxOps";
    /**
     * There was no previous emit for the last X milliseconds.
     */
    OpEventType[OpEventType["MaxTime"] = 2] = "MaxTime";
    /**
     * Indicates the durable sequence number was updated.
     */
    OpEventType[OpEventType["UpdatedDurableSequenceNumber"] = 3] = "UpdatedDurableSequenceNumber";
})(OpEventType = exports.OpEventType || (exports.OpEventType = {}));
/**
 * Check if the string is a service message type, which includes
 * MessageType.ClientJoin, MessageType.ClientLeave, MessageType.Control,
 * MessageType.NoClient, MessageType.SummaryAck, and MessageType.SummaryNack
 *
 * @param type - the type to check
 * @returns true if it is a system message type
 */
const isServiceMessageType = (type) => type === protocol_definitions_1.MessageType.ClientJoin ||
    type === protocol_definitions_1.MessageType.ClientLeave ||
    type === protocol_definitions_1.MessageType.Control ||
    type === protocol_definitions_1.MessageType.NoClient ||
    type === protocol_definitions_1.MessageType.SummaryAck ||
    type === protocol_definitions_1.MessageType.SummaryNack;
class DeliLambda extends common_utils_1.TypedEventEmitter {
    constructor(context, tenantId, documentId, lastCheckpoint, checkpointManager, clientManager, deltasProducer, signalsProducer, rawDeltasProducer, serviceConfiguration, sessionMetric, sessionStartMetric, checkpointService, restartOnCheckpointFailure, kafkaCheckpointOnReprocessingOp, sequencedSignalClients = new Map()) {
        var _a, _b, _c, _d;
        super();
        this.context = context;
        this.tenantId = tenantId;
        this.documentId = documentId;
        this.lastCheckpoint = lastCheckpoint;
        this.clientManager = clientManager;
        this.deltasProducer = deltasProducer;
        this.signalsProducer = signalsProducer;
        this.rawDeltasProducer = rawDeltasProducer;
        this.serviceConfiguration = serviceConfiguration;
        this.sessionMetric = sessionMetric;
        this.sessionStartMetric = sessionStartMetric;
        this.checkpointService = checkpointService;
        this.restartOnCheckpointFailure = restartOnCheckpointFailure;
        this.kafkaCheckpointOnReprocessingOp = kafkaCheckpointOnReprocessingOp;
        this.sequencedSignalClients = sequencedSignalClients;
        // Client sequence number mapping
        this.clientSeqManager = new clientSeqManager_1.ClientSequenceNumberManager();
        this.minimumSequenceNumber = 0;
        this.lastSendP = Promise.resolve();
        this.lastNoClientP = Promise.resolve();
        this.lastSentMSN = 0;
        this.lastInstruction = InstructionType.NoOp;
        /**
         * Used for controlling op event logic
         */
        this.opEvent = { sequencedMessagesSinceLastOpEvent: 0 };
        /**
         * Used for controlling checkpoint logic
         */
        this.checkpointInfo = {
            lastCheckpointTime: Date.now(),
            rawMessagesSinceCheckpoint: 0,
        };
        this.closed = false;
        // Session level properties
        this.serviceSummaryGenerated = false;
        this.isNewDocument = false;
        this.successfullyStartedLambdas = [];
        this.expectedSuccessfullyStartedLambdas = [server_services_core_1.LambdaName.Scribe];
        // Instantiate existing clients
        if (lastCheckpoint.clients) {
            for (const client of lastCheckpoint.clients) {
                if (client.clientId) {
                    this.clientSeqManager.upsertClient(client.clientId, client.clientSequenceNumber, client.referenceSequenceNumber, client.lastUpdate, client.canEvict, client.scopes, client.nack, client.serverMetadata);
                }
            }
        }
        // Initialize counting context
        this.sequenceNumber = lastCheckpoint.sequenceNumber;
        this.signalClientConnectionNumber = (_a = lastCheckpoint.signalClientConnectionNumber) !== null && _a !== void 0 ? _a : 0;
        this.lastHash = (_b = lastCheckpoint.expHash1) !== null && _b !== void 0 ? _b : server_services_client_1.defaultHash;
        this.durableSequenceNumber = lastCheckpoint.durableSequenceNumber;
        this.lastSentMSN = (_c = lastCheckpoint.lastSentMSN) !== null && _c !== void 0 ? _c : 0;
        this.logOffset = lastCheckpoint.logOffset;
        if (lastCheckpoint.nackMessages) {
            if (Array.isArray(lastCheckpoint.nackMessages)) {
                this.nackMessages = new Map(lastCheckpoint.nackMessages);
            }
            else {
                // backwards compat. nackMessages is a INackMessagesControlMessageContents
                this.nackMessages = new Map();
                // extra check for very old nack messages
                const identifier = lastCheckpoint.nackMessages.identifier;
                if (identifier !== undefined) {
                    this.nackMessages.set(identifier, lastCheckpoint.nackMessages);
                }
            }
        }
        else {
            this.nackMessages = new Map();
        }
        // Null coalescing for backward compatibility
        this.successfullyStartedLambdas = (_d = lastCheckpoint.successfullyStartedLambdas) !== null && _d !== void 0 ? _d : [];
        const msn = this.clientSeqManager.getMinimumSequenceNumber();
        this.noActiveClients = msn === -1;
        this.minimumSequenceNumber = this.noActiveClients ? this.sequenceNumber : msn;
        if (this.serviceConfiguration.deli.summaryNackMessages.checkOnStartup) {
            this.checkNackMessagesState();
        }
        this.checkpointContext = new checkpointContext_1.CheckpointContext(this.tenantId, this.documentId, checkpointManager, context, this.checkpointService);
        // start the activity idle timer when created
        this.setActivityIdleTimer();
        this.setReadClientIdleTimer();
        if (this.serviceConfiguration.deli.opEvent.enable) {
            this.updateOpMaxTimeTimer();
            /**
             * Deli's opEvent system is supposed to tell us when it's time to post ops for the session.
             * It sends an "opEvent" event based heuristics like idle / max time / max ops.
             * There's an edge case though. Suppose the following:
             * 1. Server A created a deli for the session, consumes 100 kafka messages, and sequences 100 ops.
             * 2. Within 5 seconds of sequencing those ops,
             * Server A's deli saves a checkpoint (it remembers it sequenced those 100 ops)
             * 3. Within a second of that checkpoint, the Kafka partition is rebalanced.
             * 4. Server B now creates a deli for that session and it consumes those same 100 kafka messages.
             * 4a. Server B's deli instance is smart enough to detect that those 100 kafka messages were already
             * processed (due to the checkpoint created in #2) so it ignores them (the first if statement in handler).
             *
             * The above flow is a problem because the opEvent logic is not going to trigger since
             * no messages were sequenced by this deli.
             *
             * Deli should be smart and check if it hasn't yet sent an opEvent for messages that
             * were not durably stored.
             */
            if (this.sequenceNumber > this.durableSequenceNumber) {
                /**
                 * This makes it so the next time deli checks for a "maxTime" opEvent,
                 * it will fire the event since sequencedMessagesSinceLastOpEvent \> 0.
                 */
                this.opEvent.sequencedMessagesSinceLastOpEvent =
                    this.sequenceNumber - this.durableSequenceNumber;
            }
        }
        this.isNewDocument = this.sequenceNumber === 0;
        if (serviceConfiguration.enableLumberjack) {
            this.logSessionStartMetrics();
        }
        if (this.serviceConfiguration.deli.checkForIdleClientsOnStartup) {
            /**
             * Instruct deli to check for idle clients on startup. Why do we want to do this?
             *
             * Suppose the following:
             * 1. Deli starts up and there is 1 write client and it
             * consumes 1 message it has already previouly consumed.
             * 2. Deli is closed due to a rebalance 2 minutes later.
             * 3. Suppose that deli keeps rebalancing every 2 minutes indefinitely.
             *
             * Deli is configured to checkpoint 1 message behind the head while there is a client in the session.
             * This will cause the kafka partition to never get a new checkpoint because it's in this bad loop.
             * Never checkpointing could eventually lead to messages expiring from Kafka (data loss/corruption).
             *
             * We can recover from this loop if we check for idle clients on startup and insert a leave message
             * for that 1 write client (who is now definitely expired). It would end up making deli checkpoint properly.
             */
            this.checkIdleWriteClients(Date.now());
        }
    }
    handler(rawMessage) {
        // In cases where we are reprocessing messages we have already checkpointed exit early
        if (this.logOffset !== undefined && rawMessage.offset <= this.logOffset) {
            const reprocessOpsMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.ReprocessOps);
            reprocessOpsMetric.setProperties(Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { kafkaMessageOffset: rawMessage.offset, databaseLastOffset: this.logOffset }));
            this.updateCheckpointMessages(rawMessage);
            try {
                if (this.checkpointInfo.currentKafkaCheckpointMessage &&
                    this.kafkaCheckpointOnReprocessingOp) {
                    this.context.checkpoint(this.checkpointInfo.currentKafkaCheckpointMessage, this.restartOnCheckpointFailure);
                }
                reprocessOpsMetric.setProperty("kafkaCheckpointOnReprocessingOp", this.kafkaCheckpointOnReprocessingOp);
                reprocessOpsMetric.success(`Successfully reprocessed repeating ops.`);
            }
            catch (error) {
                reprocessOpsMetric.error(`Error while reprocessing ops.`, error);
            }
            return undefined;
        }
        else if (this.logOffset === undefined) {
            server_services_telemetry_1.Lumberjack.error(`No value for logOffset`, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
        }
        this.logOffset = rawMessage.offset;
        let sequencedMessageCount = 0;
        // array of messages that should be produced to the deltas topic after processing
        const produceToDeltas = [];
        const boxcar = (0, server_services_core_1.extractBoxcar)(rawMessage);
        for (const message of boxcar.contents) {
            // Ticket current message.
            const ticketedMessage = this.ticket(message, this.serviceConfiguration.enableTraces ? this.createTrace("start") : undefined);
            // Return early if message is invalid
            if (!ticketedMessage) {
                continue;
            }
            this.lastInstruction = ticketedMessage.instruction;
            switch (ticketedMessage.ticketType) {
                case TicketType.Sequenced: {
                    this.lastMessageType = ticketedMessage.type;
                    if (ticketedMessage.type !== protocol_definitions_1.MessageType.ClientLeave) {
                        // Check for idle write clients.
                        this.checkIdleWriteClients(ticketedMessage.timestamp);
                    }
                    // Check for document inactivity.
                    if (!(ticketedMessage.type === protocol_definitions_1.MessageType.NoClient ||
                        ticketedMessage.type === protocol_definitions_1.MessageType.Control) &&
                        this.noActiveClients &&
                        !this.serviceConfiguration.deli.disableNoClientMessage) {
                        this.lastNoClientP = this.sendToRawDeltas(this.createOpMessage(protocol_definitions_1.MessageType.NoClient)).catch((error) => {
                            var _a;
                            const errorMsg = "Could not send no client message";
                            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`${errorMsg}: ${JSON.stringify(error)}`, {
                                messageMetaData: {
                                    documentId: this.documentId,
                                    tenantId: this.tenantId,
                                },
                            });
                            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
                            this.context.error(error, {
                                restart: true,
                                tenantId: this.tenantId,
                                documentId: this.documentId,
                            });
                        });
                    }
                    // Return early if sending is not required.
                    if (ticketedMessage.send === SendType.Never) {
                        continue;
                    }
                    // Return early but start a timer to create consolidated message.
                    this.clearNoopConsolidationTimer();
                    if (ticketedMessage.send === SendType.Later) {
                        this.setNoopConsolidationTimer();
                        continue;
                    }
                    const sequencedMessage = ticketedMessage.message;
                    if (this.serviceConfiguration.deli.enableOpHashing) {
                        this.lastHash = (0, server_services_client_1.getNextHash)(sequencedMessage, this.lastHash);
                        sequencedMessage.expHash1 = this.lastHash;
                    }
                    if (sequencedMessage.type === protocol_definitions_1.MessageType.Summarize) {
                        // note: this is being emitted before it's produced to the deltas topic
                        // that lets event handlers alter the message if necessary
                        this.emit("summarizeMessage", sequencedMessage);
                    }
                    const outgoingMessage = {
                        type: server_services_core_1.SequencedOperationType,
                        tenantId: this.tenantId,
                        documentId: this.documentId,
                        operation: sequencedMessage,
                    };
                    if (this.serviceConfiguration.deli.maintainBatches) {
                        produceToDeltas.push(outgoingMessage);
                    }
                    else {
                        this.produceMessage(this.deltasProducer, outgoingMessage);
                    }
                    sequencedMessageCount++;
                    // Update the msn last sent
                    this.lastSentMSN = ticketedMessage.msn;
                    // create a signal for a write client if all the following are true:
                    // 1. a signal producer is provided
                    // 2. the sequenced op is a join or leave message
                    // 3. enableWriteClientSignals is on or alfred told us to create a signal
                    // #3 allows alfred to be in charge of enabling this functionality
                    if (this.signalsProducer &&
                        (sequencedMessage.type === protocol_definitions_1.MessageType.ClientJoin ||
                            sequencedMessage.type === protocol_definitions_1.MessageType.ClientLeave) &&
                        (this.serviceConfiguration.deli.enableWriteClientSignals ||
                            (sequencedMessage.serverMetadata &&
                                typeof sequencedMessage.serverMetadata === "object" &&
                                sequencedMessage.serverMetadata.createSignal))) {
                        const dataContent = this.extractDataContent(message);
                        const signalMessage = this.createSignalMessage(message, sequencedMessage.sequenceNumber - 1, dataContent);
                        if (sequencedMessage.type === protocol_definitions_1.MessageType.ClientJoin) {
                            this.addSequencedSignalClient(dataContent, signalMessage);
                        }
                        else {
                            this.sequencedSignalClients.delete(dataContent);
                        }
                        this.produceMessage(this.signalsProducer, signalMessage.message);
                    }
                    break;
                }
                case TicketType.Nack: {
                    if (this.serviceConfiguration.deli.maintainBatches) {
                        produceToDeltas.push(ticketedMessage.message);
                    }
                    else {
                        this.produceMessage(this.deltasProducer, ticketedMessage.message);
                    }
                    break;
                }
                case TicketType.Signal: {
                    if (this.signalsProducer) {
                        this.produceMessage(this.signalsProducer, ticketedMessage.message);
                    }
                    break;
                }
                default:
                    // ignore unknown types
                    break;
            }
        }
        if (produceToDeltas.length > 0) {
            // processing this boxcar resulted in one or more ticketed messages (sequenced ops or nacks)
            // produce them in a single boxcar to the deltas topic
            this.produceMessages(this.deltasProducer, produceToDeltas, rawMessage);
        }
        this.checkpointInfo.rawMessagesSinceCheckpoint++;
        this.updateCheckpointMessages(rawMessage);
        if (this.lastMessageType === protocol_definitions_1.MessageType.ClientJoin) {
            this.globalCheckpointOnly = false;
        }
        const checkpointReason = this.getCheckpointReason(this.lastMessageType);
        if (checkpointReason !== undefined) {
            // Set a flag to route all checkpoints to global collection if there are no active clients, and reset when clients join
            if (checkpointReason === utils_1.CheckpointReason.NoClients) {
                this.globalCheckpointOnly = true;
            }
            // checkpoint the current up to date state
            this.checkpoint(checkpointReason, this.globalCheckpointOnly);
        }
        else {
            this.updateCheckpointIdleTimer(this.globalCheckpointOnly);
        }
        // Start a timer to check inactivity on the document. To trigger idle client leave message,
        // we send a noop back to alfred. The noop should trigger a client leave message if there are any.
        this.clearActivityIdleTimer();
        this.setActivityIdleTimer();
        if (sequencedMessageCount > 0) {
            // Check if Deli is over the max ops since last summary nack limit
            // Note: we are explicitly checking this after processing the entire boxcar in order to not break batches
            if (this.serviceConfiguration.deli.summaryNackMessages.enable &&
                !this.nackMessages.has(server_services_core_1.NackMessagesType.SummaryMaxOps)) {
                const opsSinceLastSummary = this.sequenceNumber - this.durableSequenceNumber;
                if (opsSinceLastSummary > this.serviceConfiguration.deli.summaryNackMessages.maxOps) {
                    // this op brings us over the limit
                    // start nacking non-system ops and ops that are submitted by non-summarizers
                    this.updateNackMessages(server_services_core_1.NackMessagesType.SummaryMaxOps, {
                        identifier: server_services_core_1.NackMessagesType.SummaryMaxOps,
                        content: this.serviceConfiguration.deli.summaryNackMessages.nackContent,
                        allowSystemMessages: true,
                        allowedScopes: [protocol_definitions_1.ScopeType.SummaryWrite],
                    });
                }
            }
            // Update the op event idle & max ops counter if ops were just sequenced
            if (this.serviceConfiguration.deli.opEvent.enable) {
                this.updateOpIdleTimer();
                const maxOps = this.serviceConfiguration.deli.opEvent.maxOps;
                if (maxOps !== undefined) {
                    this.opEvent.sequencedMessagesSinceLastOpEvent += sequencedMessageCount;
                    if (this.opEvent.sequencedMessagesSinceLastOpEvent > maxOps) {
                        this.emitOpEvent(OpEventType.MaxOps);
                    }
                }
            }
        }
    }
    close(closeType) {
        this.closed = true;
        this.checkpointContext.close();
        this.clearActivityIdleTimer();
        this.clearReadClientIdleTimer();
        this.clearNoopConsolidationTimer();
        this.clearCheckpointIdleTimer();
        this.clearOpIdleTimer();
        this.clearOpMaxTimeTimer();
        this.emit("close", closeType);
        this.removeAllListeners();
        if (this.serviceConfiguration.enableLumberjack) {
            this.logSessionEndMetrics(closeType);
        }
    }
    produceMessage(producer, message) {
        this.lastSendP = producer
            .send([message], message.tenantId, message.documentId)
            .catch((error) => {
            var _a;
            if (this.closed) {
                return;
            }
            const errorMsg = "Could not send message to producer";
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`${errorMsg}: ${JSON.stringify(error)}`, {
                messageMetaData: {
                    documentId: this.documentId,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
            this.context.error(error, {
                restart: true,
                tenantId: this.tenantId,
                documentId: this.documentId,
            });
        });
    }
    produceMessages(producer, messages, rawMessage) {
        this.lastSendP = producer.send(messages, this.tenantId, this.documentId).catch((error) => {
            var _a;
            if (this.closed) {
                return;
            }
            const errorMsg = `Could not send ${messages.length} messages to producer. offset: ${rawMessage.offset}`;
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`${errorMsg}: ${JSON.stringify(error)}`, {
                messageMetaData: {
                    documentId: this.documentId,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
            let restart = true;
            let markAsCorrupt = false;
            if ((0, server_services_client_1.isNetworkError)(error) && error.code === 413) {
                // kafka message size too large
                restart = false;
                markAsCorrupt = true;
            }
            this.context.error(error, {
                restart,
                markAsCorrupt: markAsCorrupt ? rawMessage : undefined,
                tenantId: this.tenantId,
                documentId: this.documentId,
            });
        });
    }
    logSessionStartMetrics(failMetric = false) {
        var _a, _b, _c, _d, _e, _f, _g, _h;
        if ((_a = this.sessionStartMetric) === null || _a === void 0 ? void 0 : _a.isCompleted()) {
            this.sessionStartMetric = (0, utils_1.createSessionMetric)(this.tenantId, this.documentId, server_services_telemetry_1.LumberEventName.StartSessionResult, this.serviceConfiguration);
        }
        if (failMetric) {
            (_b = this.sessionStartMetric) === null || _b === void 0 ? void 0 : _b.setProperties({
                [server_services_telemetry_1.CommonProperties.sessionState]: server_services_telemetry_1.SessionState.LambdaStartFailed,
            });
            (_c = this.sessionStartMetric) === null || _c === void 0 ? void 0 : _c.error("Lambda start failed");
            return;
        }
        if (this.verifyRequiredLambdaStarted()) {
            if (this.isNewDocument) {
                (_d = this.sessionStartMetric) === null || _d === void 0 ? void 0 : _d.setProperties({
                    [server_services_telemetry_1.CommonProperties.sessionState]: server_services_telemetry_1.SessionState.started,
                });
                (_e = this.sessionStartMetric) === null || _e === void 0 ? void 0 : _e.success("Session started successfully");
            }
            else {
                (_f = this.sessionStartMetric) === null || _f === void 0 ? void 0 : _f.setProperties({
                    [server_services_telemetry_1.CommonProperties.sessionState]: server_services_telemetry_1.SessionState.resumed,
                });
                (_g = this.sessionStartMetric) === null || _g === void 0 ? void 0 : _g.success("Session resumed successfully");
            }
        }
        else {
            const lambdaStatusMsg = "Not all required lambdas started";
            (_h = this.context.log) === null || _h === void 0 ? void 0 : _h.info(lambdaStatusMsg);
            server_services_telemetry_1.Lumberjack.info(lambdaStatusMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
        }
    }
    verifyRequiredLambdaStarted() {
        return this.expectedSuccessfullyStartedLambdas.every((val) => this.successfullyStartedLambdas.includes(val));
    }
    logSessionEndMetrics(closeType) {
        var _a, _b;
        if ((_a = this.sessionMetric) === null || _a === void 0 ? void 0 : _a.isCompleted()) {
            server_services_telemetry_1.Lumberjack.info("Session metric already completed. Creating a new one.", (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
            this.sessionMetric = (0, utils_1.createSessionMetric)(this.tenantId, this.documentId, server_services_telemetry_1.LumberEventName.SessionResult, this.serviceConfiguration);
        }
        (_b = this.sessionMetric) === null || _b === void 0 ? void 0 : _b.setProperties({
            [server_services_telemetry_1.CommonProperties.serviceSummarySuccess]: this.serviceSummaryGenerated,
        });
        (0, utils_1.logCommonSessionEndMetrics)(this.context, closeType, this.sessionMetric, this.sequenceNumber, this.durableSequenceNumber, Array.from(this.nackMessages.keys()));
    }
    ticket(rawMessage, trace) {
        var _a, _b, _c, _d, _e;
        var _f;
        // Exit out early for unknown messages
        if (rawMessage.type !== server_services_core_1.RawOperationType) {
            return undefined;
        }
        // Update and retrieve the minimum sequence number
        const message = rawMessage;
        const dataContent = this.extractDataContent(message);
        // Check if we should nack this message
        if (this.nackMessages.size > 0 && this.serviceConfiguration.deli.enableNackMessages) {
            for (const nackMessageControlMessageContents of this.nackMessages.values()) {
                let shouldNack = true;
                if (nackMessageControlMessageContents.allowSystemMessages &&
                    (isServiceMessageType(message.operation.type) || !message.clientId)) {
                    // this is a system message. don't nack it
                    shouldNack = false;
                }
                else if (nackMessageControlMessageContents.allowedScopes) {
                    const clientId = message.clientId;
                    if (clientId) {
                        const client = this.clientSeqManager.get(clientId);
                        if (client) {
                            for (const scope of nackMessageControlMessageContents.allowedScopes) {
                                if (client.scopes.includes(scope)) {
                                    // this client has an allowed scope. don't nack it
                                    shouldNack = false;
                                    break;
                                }
                            }
                        }
                    }
                }
                if (shouldNack) {
                    return this.createNackMessage(message, nackMessageControlMessageContents.content.code, nackMessageControlMessageContents.content.type, nackMessageControlMessageContents.content.message, nackMessageControlMessageContents.content.retryAfter);
                }
            }
        }
        // Check incoming message order. Nack if there is any gap so that the client can resend.
        const messageOrder = this.checkOrder(message);
        if (messageOrder === IncomingMessageOrder.Duplicate) {
            return;
        }
        else if (messageOrder === IncomingMessageOrder.Gap) {
            return this.createNackMessage(message, 400, protocol_definitions_1.NackErrorType.BadRequestError, `Gap detected in incoming op`);
        }
        if (this.isInvalidMessage(message)) {
            return this.createNackMessage(message, 400, protocol_definitions_1.NackErrorType.BadRequestError, `Op not allowed`);
        }
        // Handle client join/leave messages.
        if (!message.clientId) {
            if (message.operation.type === protocol_definitions_1.MessageType.ClientLeave) {
                if (!this.clientSeqManager.removeClient(dataContent)) {
                    // not a write client. check if it was a read client
                    const readClient = this.sequencedSignalClients.get(dataContent);
                    if (readClient) {
                        this.sequencedSignalClients.delete(dataContent);
                        return this.createSignalMessage(message, this.sequenceNumber, dataContent);
                    }
                    // Return if the client has already been removed due to a prior leave message.
                    return;
                }
                if (this.serviceConfiguration.deli.enableLeaveOpNoClientServerMetadata &&
                    this.clientSeqManager.count() === 0) {
                    // add server metadata to indicate the last client left
                    (_a = (_f = message.operation).serverMetadata) !== null && _a !== void 0 ? _a : (_f.serverMetadata = {});
                    message.operation.serverMetadata.noClient = true;
                }
            }
            else if (message.operation.type === protocol_definitions_1.MessageType.ClientJoin) {
                const clientJoinMessage = dataContent;
                if (clientJoinMessage.detail.mode === "read") {
                    if (this.sequencedSignalClients.has(clientJoinMessage.clientId)) {
                        // Return if the client has already been added due to a prior join message.
                        return;
                    }
                    // create the signal message
                    const signalMessage = this.createSignalMessage(message, this.sequenceNumber, dataContent);
                    this.addSequencedSignalClient(clientJoinMessage, signalMessage);
                    return signalMessage;
                }
                else {
                    const isNewClient = this.clientSeqManager.upsertClient(clientJoinMessage.clientId, 0, this.minimumSequenceNumber, message.timestamp, true, clientJoinMessage.detail.scopes, false, message.operation.serverMetadata);
                    if (!isNewClient) {
                        // Return if the client has already been added due to a prior join message.
                        return;
                    }
                }
            }
        }
        else {
            // Nack inexistent client.
            const client = this.clientSeqManager.get(message.clientId);
            if (!client || client.nack) {
                return this.createNackMessage(message, 400, protocol_definitions_1.NackErrorType.BadRequestError, `Nonexistent client`);
            }
            // Verify that the message is within the current window.
            // -1 check just for directly sent ops (e.g., using REST API).
            if (message.clientId &&
                message.operation.referenceSequenceNumber !== -1 &&
                message.operation.referenceSequenceNumber < this.minimumSequenceNumber) {
                this.clientSeqManager.upsertClient(message.clientId, message.operation.clientSequenceNumber, this.minimumSequenceNumber, message.timestamp, true, [], true);
                return this.createNackMessage(message, 400, protocol_definitions_1.NackErrorType.BadRequestError, `Refseq ${message.operation.referenceSequenceNumber} < ${this.minimumSequenceNumber}`);
            }
            // Nack if an unauthorized client tries to summarize.
            if (message.operation.type === protocol_definitions_1.MessageType.Summarize) {
                if (!(0, server_services_client_1.canSummarize)(client.scopes)) {
                    return this.createNackMessage(message, 403, protocol_definitions_1.NackErrorType.InvalidScopeError, `Client ${message.clientId} does not have summary permission`);
                }
            }
        }
        let sequenceNumber = this.sequenceNumber;
        // Get the current sequence number and increment it if appropriate.
        // We don't increment sequence number for noops sent by client since they will
        // be consolidated and sent later as raw message.
        if (message.clientId) {
            // Don't rev for client sent no-ops
            if (message.operation.type !== protocol_definitions_1.MessageType.NoOp) {
                // Rev the sequence number
                sequenceNumber = this.revSequenceNumber();
            }
            // Only for directly sent ops (e.g., using REST API). To avoid getting nacked,
            // We rev the refseq number to current sequence number.
            if (message.operation.referenceSequenceNumber === -1) {
                message.operation.referenceSequenceNumber = sequenceNumber;
            }
            this.clientSeqManager.upsertClient(message.clientId, message.operation.clientSequenceNumber, message.operation.referenceSequenceNumber, message.timestamp, true);
        }
        else {
            // Don't rev for server sent no-ops, noClient, or Control messages.
            if (!(message.operation.type === protocol_definitions_1.MessageType.NoOp ||
                message.operation.type === protocol_definitions_1.MessageType.NoClient ||
                message.operation.type === protocol_definitions_1.MessageType.Control)) {
                sequenceNumber = this.revSequenceNumber();
            }
        }
        // Store the previous minimum sequence number we returned and then update it. If there are no clients
        // then set the MSN to the next SN.
        const msn = this.clientSeqManager.getMinimumSequenceNumber();
        if (msn === -1) {
            this.minimumSequenceNumber = sequenceNumber;
            this.noActiveClients = true;
        }
        else {
            this.minimumSequenceNumber = msn;
            this.noActiveClients = false;
        }
        let sendType = SendType.Immediate;
        let instruction = InstructionType.NoOp;
        /**
         * Run extra logic depending on the op type
         */
        switch (message.operation.type) {
            /**
             * Sequence number was never rev'd for NoOps. We will decide now based on heuristics.
             */
            case protocol_definitions_1.MessageType.NoOp: {
                // Set up delay sending of client sent no-ops
                if (message.clientId) {
                    if (message.operation.contents === null) {
                        sendType = SendType.Later;
                    }
                    else {
                        if (this.minimumSequenceNumber <= this.lastSentMSN) {
                            sendType = SendType.Later;
                        }
                        else {
                            sequenceNumber = this.revSequenceNumber();
                        }
                    }
                }
                else {
                    if (this.minimumSequenceNumber <= this.lastSentMSN) {
                        sendType = SendType.Never;
                    }
                    else {
                        // Only rev if we need to send a new msn.
                        sequenceNumber = this.revSequenceNumber();
                    }
                }
                break;
            }
            /**
             * Sequence number was never rev'd for noClients. We will decide now based on heuristics.
             */
            case protocol_definitions_1.MessageType.NoClient: {
                // Only rev if no clients have shown up since last noClient was sent to alfred.
                if (this.noActiveClients) {
                    sequenceNumber = this.revSequenceNumber();
                    message.operation.referenceSequenceNumber = sequenceNumber;
                    this.minimumSequenceNumber = sequenceNumber;
                }
                else {
                    sendType = SendType.Never;
                }
                break;
            }
            case protocol_definitions_1.MessageType.Control: {
                sendType = SendType.Never;
                const controlMessage = dataContent;
                switch (controlMessage.type) {
                    case server_services_core_1.ControlMessageType.UpdateDSN: {
                        const dsnStatusMsg = `Update DSN: ${JSON.stringify(controlMessage)}`;
                        (_b = this.context.log) === null || _b === void 0 ? void 0 : _b.info(dsnStatusMsg, {
                            messageMetaData: {
                                documentId: this.documentId,
                                tenantId: this.tenantId,
                            },
                        });
                        server_services_telemetry_1.Lumberjack.info(dsnStatusMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                        const controlContents = controlMessage.contents;
                        this.serviceSummaryGenerated = !controlContents.isClientSummary;
                        const dsn = controlContents.durableSequenceNumber;
                        if (dsn >= this.durableSequenceNumber) {
                            // Deli cache is only cleared when no clients have
                            // joined since last noClient was sent to alfred
                            if (controlContents.clearCache && this.noActiveClients) {
                                instruction = InstructionType.ClearCache;
                                const deliCacheMsg = `Deli cache will be cleared`;
                                (_c = this.context.log) === null || _c === void 0 ? void 0 : _c.info(deliCacheMsg, {
                                    messageMetaData: {
                                        documentId: this.documentId,
                                        tenantId: this.tenantId,
                                    },
                                });
                                server_services_telemetry_1.Lumberjack.info(deliCacheMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                            }
                            this.updateDurableSequenceNumber(dsn);
                        }
                        break;
                    }
                    case server_services_core_1.ControlMessageType.NackMessages: {
                        const controlContents = controlMessage.contents;
                        this.updateNackMessages(controlContents.identifier, controlContents.content !== undefined ? controlContents : undefined);
                        break;
                    }
                    case server_services_core_1.ControlMessageType.LambdaStartResult: {
                        const controlContents = controlMessage.contents;
                        if (controlContents.success) {
                            this.successfullyStartedLambdas.push(controlContents.lambdaName);
                        }
                        this.logSessionStartMetrics(!controlContents.success);
                        break;
                    }
                    case server_services_core_1.ControlMessageType.ExtendClient: {
                        const controlContents = controlMessage.contents;
                        const clientsToExtend = new Map();
                        const clientIds = (_d = controlContents.clientIds) !== null && _d !== void 0 ? _d : (controlContents.clientId ? [controlContents.clientId] : []);
                        for (const clientId of clientIds) {
                            const client = this.sequencedSignalClients.get(clientId);
                            if (client) {
                                clientsToExtend.set(clientId, client);
                            }
                        }
                        if (clientsToExtend.size > 0) {
                            if (this.clientManager) {
                                this.clientManager
                                    .extendSequencedClients(this.tenantId, this.documentId, clientsToExtend, this.serviceConfiguration.deli.clientTimeout)
                                    .catch((error) => {
                                    var _a;
                                    const errorMsg = "Could not extend clients";
                                    (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`${errorMsg}: ${JSON.stringify(error)}`, {
                                        messageMetaData: {
                                            documentId: this.documentId,
                                            tenantId: this.tenantId,
                                        },
                                    });
                                    server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
                                });
                            }
                            else {
                                const errorMsg = "Could not extend clients. Missing client manager";
                                (_e = this.context.log) === null || _e === void 0 ? void 0 : _e.error(`${errorMsg}`, {
                                    messageMetaData: {
                                        documentId: this.documentId,
                                        tenantId: this.tenantId,
                                    },
                                });
                                server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
                            }
                        }
                        break;
                    }
                    default:
                        // an unknown control message was received
                        // emit a control message event to support custom control messages
                        this.emit("controlMessage", controlMessage);
                        break;
                }
                break;
            }
            /**
             * Automatically update the DSN when sequencing a summaryAck
             */
            case protocol_definitions_1.MessageType.SummaryAck: {
                if (this.serviceConfiguration.deli.enableAutoDSNUpdate) {
                    const dsn = dataContent.summaryProposal.summarySequenceNumber;
                    if (dsn >= this.durableSequenceNumber) {
                        this.updateDurableSequenceNumber(dsn);
                    }
                }
                break;
            }
            default:
                break;
        }
        // Add traces
        if (trace && message.operation.traces && message.operation.traces.length > 1) {
            message.operation.traces.push(trace);
            message.operation.traces.push(this.createTrace("end"));
        }
        // craft the output message
        const outputMessage = this.createOutputMessage(message, undefined /* origin */, sequenceNumber, dataContent);
        return {
            ticketType: TicketType.Sequenced,
            instruction,
            message: outputMessage,
            msn: this.minimumSequenceNumber,
            send: sendType,
            timestamp: message.timestamp,
            type: message.operation.type,
        };
    }
    extractDataContent(message) {
        if (message.operation.type === protocol_definitions_1.MessageType.ClientJoin ||
            message.operation.type === protocol_definitions_1.MessageType.ClientLeave ||
            message.operation.type === protocol_definitions_1.MessageType.SummaryAck ||
            message.operation.type === protocol_definitions_1.MessageType.SummaryNack ||
            message.operation.type === protocol_definitions_1.MessageType.Control) {
            const operation = message.operation;
            if (operation.data) {
                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return JSON.parse(operation.data);
            }
        }
    }
    isInvalidMessage(message) {
        return message.clientId ? isServiceMessageType(message.operation.type) : false;
    }
    createOutputMessage(message, origin, sequenceNumber, dataContent) {
        var _a;
        const outputMessage = {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            clientId: message.clientId,
            clientSequenceNumber: message.operation.clientSequenceNumber,
            contents: message.operation.contents,
            metadata: message.operation.metadata,
            serverMetadata: message.operation.serverMetadata,
            minimumSequenceNumber: this.minimumSequenceNumber,
            origin,
            referenceSequenceNumber: message.operation.referenceSequenceNumber,
            sequenceNumber,
            // "term" was an experimental feature that is being removed.  The only safe value to use is 1.
            term: 1,
            timestamp: message.timestamp,
            traces: message.operation.traces,
            type: message.operation.type,
            compression: message.operation.compression,
        };
        if (message.operation.type === protocol_definitions_1.MessageType.Summarize ||
            message.operation.type === protocol_definitions_1.MessageType.NoClient) {
            const augmentedOutputMessage = outputMessage;
            // only add additional content if scribe will use this op for generating a summary
            // NoClient ops are ignored by scribe when generateServiceSummary is disabled
            let addAdditionalContent = false;
            if (this.serviceConfiguration.scribe.generateServiceSummary) {
                addAdditionalContent = true;
            }
            else if (message.operation.type === protocol_definitions_1.MessageType.Summarize) {
                // no need to add additionalContent for summarize messages using the single commit flow
                // because scribe will not be involved
                if (!this.serviceConfiguration.deli.skipSummarizeAugmentationForSingleCommmit ||
                    !((_a = JSON.parse(message.operation.contents).details) === null || _a === void 0 ? void 0 : _a.includesProtocolTree)) {
                    addAdditionalContent = true;
                }
            }
            if (addAdditionalContent) {
                const checkpointData = JSON.stringify(this.generateDeliCheckpoint());
                augmentedOutputMessage.additionalContent = checkpointData;
            }
            return augmentedOutputMessage;
        }
        else if (dataContent !== undefined) {
            // TODO to consolidate the logic here
            const systemOutputMessage = outputMessage;
            systemOutputMessage.data = JSON.stringify(dataContent);
            return systemOutputMessage;
        }
        else {
            return outputMessage;
        }
    }
    checkOrder(message) {
        var _a, _b;
        if (!message.clientId) {
            return IncomingMessageOrder.ConsecutiveOrSystem;
        }
        const clientId = message.clientId;
        const clientSequenceNumber = message.operation.clientSequenceNumber;
        const client = this.clientSeqManager.get(clientId);
        if (!client) {
            return IncomingMessageOrder.ConsecutiveOrSystem;
        }
        const messageMetaData = {
            documentId: this.documentId,
            tenantId: this.tenantId,
        };
        // Perform duplicate and gap detection - Check that we have a monotonically increasing CID
        const expectedClientSequenceNumber = client.clientSequenceNumber + 1;
        if (clientSequenceNumber === expectedClientSequenceNumber) {
            return IncomingMessageOrder.ConsecutiveOrSystem;
        }
        else if (clientSequenceNumber > expectedClientSequenceNumber) {
            const gapDetectionMsg = `Gap ${clientId}:${expectedClientSequenceNumber} > ${clientSequenceNumber}`;
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.info(gapDetectionMsg, { messageMetaData });
            server_services_telemetry_1.Lumberjack.info(gapDetectionMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
            return IncomingMessageOrder.Gap;
        }
        else {
            const dupDetectionMsg = `Duplicate ${clientId}:${expectedClientSequenceNumber} < ${clientSequenceNumber}`;
            (_b = this.context.log) === null || _b === void 0 ? void 0 : _b.info(dupDetectionMsg, { messageMetaData });
            server_services_telemetry_1.Lumberjack.info(dupDetectionMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
            return IncomingMessageOrder.Duplicate;
        }
    }
    /**
     * Sends a message to the rawdeltas queue.
     * This essentially sends the message to this deli lambda
     */
    async sendToRawDeltas(message) {
        var _a;
        try {
            await this.rawDeltasProducer.send([message], message.tenantId, message.documentId);
        }
        catch (error) {
            if (this.closed) {
                return;
            }
            const errorMsg = `Could not send message to alfred`;
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`${errorMsg}: ${JSON.stringify(error)}`, {
                messageMetaData: {
                    documentId: this.documentId,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
            this.context.error(error, {
                restart: true,
                tenantId: this.tenantId,
                documentId: this.documentId,
            });
        }
    }
    /**
     * Check if there are any old/idle write clients.
     * Craft and send a leave message if one is found.
     * To prevent recurrent leave message sending, leave messages are only piggybacked with other message type.
     */
    checkIdleWriteClients(timestamp) {
        const idleClient = this.getIdleClient(timestamp);
        if (idleClient === null || idleClient === void 0 ? void 0 : idleClient.clientId) {
            const leaveMessage = this.createLeaveMessage(idleClient.clientId, idleClient.serverMetadata);
            void this.sendToRawDeltas(leaveMessage);
        }
    }
    /**
     * Check if there are any expired read clients.
     * The read client will expire if alfred has not sent
     * an ExtendClient control message within the time for 'clientTimeout'.
     * Craft and send a leave message for each one found.
     */
    checkIdleReadClients() {
        const currentTime = Date.now();
        for (const [clientId, { client, exp }] of this.sequencedSignalClients) {
            // only handle read clients here
            // write client idle is handled by checkIdleWriteClients
            if (client.mode === "read" && exp < currentTime) {
                const leaveMessage = this.createLeaveMessage(clientId);
                void this.sendToRawDeltas(leaveMessage);
            }
        }
    }
    /**
     * Creates a leave message for inactive clients.
     */
    createLeaveMessage(clientId, serverMetadata) {
        const leaveMessage = {
            clientSequenceNumber: -1,
            contents: null,
            data: JSON.stringify(clientId),
            referenceSequenceNumber: -1,
            traces: this.serviceConfiguration.enableTraces ? [] : undefined,
            type: protocol_definitions_1.MessageType.ClientLeave,
            serverMetadata,
        };
        return this.createRawOperationMessage(leaveMessage);
    }
    /**
     * Creates a nack message for clients.
     */
    createNackMessage(message, code, type, reason, retryAfter) {
        const clientId = message.clientId;
        if (!clientId) {
            // message was sent by the system and not a client
            // "nacking" the system is not supported
            return undefined;
        }
        const nackMessage = {
            clientId,
            documentId: this.documentId,
            operation: {
                content: {
                    code,
                    type,
                    message: reason,
                    retryAfter,
                },
                operation: message.operation,
                sequenceNumber: this.minimumSequenceNumber,
            },
            tenantId: this.tenantId,
            timestamp: Date.now(),
            type: server_services_core_1.NackOperationType,
        };
        return {
            ticketType: TicketType.Nack,
            message: nackMessage,
        };
    }
    /**
     * Creates a signal message for clients.
     */
    createSignalMessage(message, sequenceNumber, dataContent) {
        let signalMessage;
        switch (message.operation.type) {
            case protocol_definitions_1.MessageType.ClientJoin:
                signalMessage = (0, utils_1.createRoomJoinMessage)(dataContent.clientId, dataContent.detail);
                break;
            case protocol_definitions_1.MessageType.ClientLeave:
                signalMessage = (0, utils_1.createRoomLeaveMessage)(typeof dataContent === "string" ? dataContent : dataContent.clientId);
                break;
            case protocol_definitions_1.MessageType.Control:
                // this will tell broadcaster to process the control message the client
                signalMessage = {
                    clientId: null,
                    content: JSON.stringify({
                        type: protocol_definitions_1.MessageType.Control,
                        content: dataContent,
                    }),
                };
                break;
            default:
                throw new Error(`Cannot create signal message for type ${message.operation.type}`);
        }
        signalMessage.referenceSequenceNumber = sequenceNumber;
        signalMessage.clientConnectionNumber = ++this.signalClientConnectionNumber;
        return {
            ticketType: TicketType.Signal,
            message: {
                type: server_services_core_1.SignalOperationType,
                tenantId: this.tenantId,
                documentId: this.documentId,
                operation: signalMessage,
                timestamp: Date.now(),
            },
        };
    }
    createOpMessage(type) {
        return this.createRawOperationMessage({
            clientSequenceNumber: -1,
            contents: null,
            referenceSequenceNumber: -1,
            traces: this.serviceConfiguration.enableTraces ? [] : undefined,
            type,
        });
    }
    createRawOperationMessage(operation) {
        return {
            clientId: null,
            documentId: this.documentId,
            operation,
            tenantId: this.tenantId,
            timestamp: Date.now(),
            type: server_services_core_1.RawOperationType,
        };
    }
    /**
     * Creates a new trace
     */
    createTrace(action) {
        const trace = {
            action,
            service: "deli",
            timestamp: Date.now(),
        };
        return trace;
    }
    /**
     * The deli checkpoint is based on rawMessage
     * The kafka checkpoint is based on kafkaCheckpointMessage if clients exist
     * This keeps the kafka checkpoint behind by 1 message until there are no active clients
     * It ensures that the idle timer and subsequent leave & NoClient messages are created
     * If noActiveClients is set, that means we sent a NoClient message. so checkpoint the current offset
     */
    updateCheckpointMessages(rawMessage) {
        this.checkpointInfo.currentCheckpointMessage = rawMessage;
        if (this.noActiveClients) {
            // If noActiveClients is set, that means we sent a NoClient message
            // so we should checkpoint the current message/offset
            // we need to explicitly set nextKafkaCheckpointMessage to undefined!
            // because once we checkpoint the current message, DocumentContext.hasPendingWork() will be false
            // that means that the partition will keep checkpointing since this lambda is up to date
            // if we don't clear nextKafkaCheckpointMessage,
            // it will try to checkpoint that old message offset once the next message arrives
            this.checkpointInfo.nextKafkaCheckpointMessage = undefined;
            this.checkpointInfo.currentKafkaCheckpointMessage = rawMessage;
        }
        else {
            // Keep the kafka checkpoint behind by 1 message until there are no active clients
            const kafkaCheckpointMessage = this.checkpointInfo.nextKafkaCheckpointMessage;
            this.checkpointInfo.nextKafkaCheckpointMessage = rawMessage;
            this.checkpointInfo.currentKafkaCheckpointMessage = kafkaCheckpointMessage;
        }
    }
    /**
     * Generates a checkpoint for the current state
     */
    generateCheckpoint(reason) {
        return {
            reason,
            deliState: this.generateDeliCheckpoint(),
            deliCheckpointMessage: this.checkpointInfo.currentCheckpointMessage,
            kafkaCheckpointMessage: this.checkpointInfo.currentKafkaCheckpointMessage,
        };
    }
    generateDeliCheckpoint() {
        return {
            clients: this.clientSeqManager.cloneValues(),
            durableSequenceNumber: this.durableSequenceNumber,
            expHash1: this.lastHash,
            logOffset: this.logOffset,
            sequenceNumber: this.sequenceNumber,
            signalClientConnectionNumber: this.signalClientConnectionNumber,
            lastSentMSN: this.lastSentMSN,
            nackMessages: Array.from(this.nackMessages),
            successfullyStartedLambdas: this.successfullyStartedLambdas,
            checkpointTimestamp: Date.now(),
        };
    }
    /**
     * Returns a new sequence number
     */
    revSequenceNumber() {
        return ++this.sequenceNumber;
    }
    /**
     * Get idle client.
     */
    getIdleClient(timestamp) {
        const client = this.clientSeqManager.peek();
        if ((client === null || client === void 0 ? void 0 : client.canEvict) &&
            timestamp - client.lastUpdate > this.serviceConfiguration.deli.clientTimeout) {
            return client;
        }
    }
    setActivityIdleTimer() {
        if (this.noActiveClients) {
            return;
        }
        this.activityIdleTimer = setTimeout(() => {
            if (!this.noActiveClients) {
                const noOpMessage = this.createOpMessage(protocol_definitions_1.MessageType.NoOp);
                void this.sendToRawDeltas(noOpMessage);
            }
        }, this.serviceConfiguration.deli.activityTimeout);
    }
    clearActivityIdleTimer() {
        if (this.activityIdleTimer !== undefined) {
            clearTimeout(this.activityIdleTimer);
            this.activityIdleTimer = undefined;
        }
    }
    setReadClientIdleTimer() {
        this.clearReadClientIdleTimer();
        this.readClientIdleTimer = setInterval(() => {
            this.checkIdleReadClients();
        }, this.serviceConfiguration.deli.readClientIdleTimer);
    }
    clearReadClientIdleTimer() {
        if (this.readClientIdleTimer !== undefined) {
            clearInterval(this.readClientIdleTimer);
            this.readClientIdleTimer = undefined;
        }
    }
    setNoopConsolidationTimer() {
        if (this.noActiveClients) {
            return;
        }
        this.noopEvent = setTimeout(() => {
            if (!this.noActiveClients) {
                const noOpMessage = this.createOpMessage(protocol_definitions_1.MessageType.NoOp);
                void this.sendToRawDeltas(noOpMessage);
            }
        }, this.serviceConfiguration.deli.noOpConsolidationTimeout);
    }
    clearNoopConsolidationTimer() {
        if (this.noopEvent !== undefined) {
            clearTimeout(this.noopEvent);
            this.noopEvent = undefined;
        }
    }
    /**
     * Reset the op event idle timer
     * Called after a message is sequenced
     */
    updateOpIdleTimer() {
        const idleTime = this.serviceConfiguration.deli.opEvent.idleTime;
        if (idleTime === undefined) {
            return;
        }
        this.clearOpIdleTimer();
        this.opEvent.idleTimer = setTimeout(() => {
            this.emitOpEvent(OpEventType.Idle);
        }, idleTime);
    }
    clearOpIdleTimer() {
        if (this.opEvent.idleTimer !== undefined) {
            clearTimeout(this.opEvent.idleTimer);
            this.opEvent.idleTimer = undefined;
        }
    }
    /**
     * Resets the op event MaxTime timer
     * Called after an opEvent is emitted
     */
    updateOpMaxTimeTimer() {
        const maxTime = this.serviceConfiguration.deli.opEvent.maxTime;
        if (maxTime === undefined) {
            return;
        }
        this.clearOpMaxTimeTimer();
        this.opEvent.maxTimer = setTimeout(() => {
            this.emitOpEvent(OpEventType.MaxTime);
        }, maxTime);
    }
    clearOpMaxTimeTimer() {
        if (this.opEvent.maxTimer !== undefined) {
            clearTimeout(this.opEvent.maxTimer);
            this.opEvent.maxTimer = undefined;
        }
    }
    /**
     * Emits an opEvent for the provided type
     * Also resets the MaxTime timer
     */
    emitOpEvent(type, force) {
        if (!force && this.opEvent.sequencedMessagesSinceLastOpEvent === 0) {
            // no need to emit since no messages were handled since last time
            return;
        }
        this.emit("opEvent", type, this.sequenceNumber, this.opEvent.sequencedMessagesSinceLastOpEvent);
        this.opEvent.sequencedMessagesSinceLastOpEvent = 0;
        this.updateOpMaxTimeTimer();
    }
    /**
     * Checks if the nackMessages flag should be reset
     */
    checkNackMessagesState() {
        if (this.serviceConfiguration.deli.summaryNackMessages.enable &&
            this.nackMessages.has(server_services_core_1.NackMessagesType.SummaryMaxOps)) {
            // Deli is nacking messages due to summary max ops
            // Check if this new dsn gets it out of that state
            const opsSinceLastSummary = this.sequenceNumber - this.durableSequenceNumber;
            if (opsSinceLastSummary <= this.serviceConfiguration.deli.summaryNackMessages.maxOps) {
                // stop nacking future messages
                this.updateNackMessages(server_services_core_1.NackMessagesType.SummaryMaxOps, undefined);
            }
        }
    }
    /**
     * Determines a checkpoint reason based on some heuristics
     * @returns a reason when it's time to checkpoint, or undefined if no checkpoint should be made
     */
    getCheckpointReason(messageType) {
        const checkpointHeuristics = this.serviceConfiguration.deli.checkpointHeuristics;
        if (!checkpointHeuristics.enable) {
            // always checkpoint since heuristics are disabled
            return utils_1.CheckpointReason.EveryMessage;
        }
        if (this.checkpointInfo.rawMessagesSinceCheckpoint >= checkpointHeuristics.maxMessages) {
            // exceeded max messages since last checkpoint
            return utils_1.CheckpointReason.MaxMessages;
        }
        if (Date.now() - this.checkpointInfo.lastCheckpointTime >= checkpointHeuristics.maxTime) {
            // exceeded max time since last checkpoint
            return utils_1.CheckpointReason.MaxTime;
        }
        if (this.lastInstruction === InstructionType.ClearCache) {
            // last instruction is for clearing the cache
            // checkpoint now to ensure that happens
            return utils_1.CheckpointReason.ClearCache;
        }
        if (this.noActiveClients && messageType === protocol_definitions_1.MessageType.NoClient) {
            return utils_1.CheckpointReason.NoClients;
        }
        return undefined;
    }
    /**
     * Checkpoints the current state once the pending kafka messages are produced
     */
    checkpoint(reason, globalCheckpointOnly) {
        this.clearCheckpointIdleTimer();
        this.checkpointInfo.lastCheckpointTime = Date.now();
        this.checkpointInfo.rawMessagesSinceCheckpoint = 0;
        Promise.all([this.lastSendP, this.lastNoClientP])
            .then(() => {
            var _a, _b;
            const checkpointParams = this.generateCheckpoint(reason);
            if (reason === utils_1.CheckpointReason.ClearCache) {
                checkpointParams.clear = true;
            }
            void this.checkpointContext.checkpoint(checkpointParams, this.restartOnCheckpointFailure, globalCheckpointOnly);
            const checkpointReason = utils_1.CheckpointReason[checkpointParams.reason];
            const checkpointResult = `Writing checkpoint. Reason: ${checkpointReason}`;
            const lumberjackProperties = Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { checkpointReason, lastOffset: this.logOffset, deliCheckpointOffset: checkpointParams.deliCheckpointMessage.offset, deliCheckpointPartition: checkpointParams.deliCheckpointMessage.partition, kafkaCheckpointOffset: (_a = checkpointParams.kafkaCheckpointMessage) === null || _a === void 0 ? void 0 : _a.offset, kafkaCheckpointPartition: (_b = checkpointParams.kafkaCheckpointMessage) === null || _b === void 0 ? void 0 : _b.partition });
            server_services_telemetry_1.Lumberjack.info(checkpointResult, lumberjackProperties);
        })
            .catch((error) => {
            var _a;
            const errorMsg = `Could not send message to scriptorium`;
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.error(`${errorMsg}: ${JSON.stringify(error)}`, {
                messageMetaData: {
                    documentId: this.documentId,
                    tenantId: this.tenantId,
                },
            });
            server_services_telemetry_1.Lumberjack.error(errorMsg, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), error);
            this.context.error(error, {
                restart: true,
                tenantId: this.tenantId,
                documentId: this.documentId,
            });
        });
    }
    /**
     * Updates the time until the state is checkpointed when idle
     * @param rawMessage - The current raw message that is initiating the timer
     */
    updateCheckpointIdleTimer(globalCheckpointOnly = false) {
        this.clearCheckpointIdleTimer();
        const initialDeliCheckpointMessage = this.checkpointInfo.currentCheckpointMessage;
        this.checkpointInfo.idleTimer = setTimeout(() => {
            this.checkpointInfo.idleTimer = undefined;
            // verify that the current deli message matches the raw message that kicked off this timer
            // if it matches, that means that delis state is for the raw message
            // this means our checkpoint will result in the correct state
            if (initialDeliCheckpointMessage === this.checkpointInfo.currentCheckpointMessage) {
                this.checkpoint(utils_1.CheckpointReason.IdleTime, globalCheckpointOnly);
            }
        }, this.serviceConfiguration.deli.checkpointHeuristics.idleTime);
    }
    /**
     * Clears the timer used for checkpointing when deli is idle
     */
    clearCheckpointIdleTimer() {
        if (this.checkpointInfo.idleTimer !== undefined) {
            clearTimeout(this.checkpointInfo.idleTimer);
            this.checkpointInfo.idleTimer = undefined;
        }
    }
    /**
     * Updates the durable sequence number
     * @param dsn - New durable sequence number
     */
    updateDurableSequenceNumber(dsn) {
        this.durableSequenceNumber = dsn;
        this.checkNackMessagesState();
        this.emit("updatedDurableSequenceNumber", dsn);
        if (this.serviceConfiguration.deli.opEvent.enable) {
            // ops were reliably stored
            // ensure op event timers & last sequenced op counters are reset
            // that will make the MaxTime & MaxOps op events accurate
            this.emitOpEvent(OpEventType.UpdatedDurableSequenceNumber, true);
        }
    }
    /**
     * Adds/updates/removes a nack message
     * @param type - Nack message type
     * @param contents - Nack messages contents or undefined to delete the nack message
     */
    updateNackMessages(type, contents) {
        if (contents !== undefined) {
            this.nackMessages.set(type, contents);
        }
        else {
            this.nackMessages.delete(type);
        }
        this.emit("updatedNackMessages", type, contents);
    }
    /**
     * Adds a sequenced signal client to the in-memory map.
     * Alfred will periodically send ExtendClient control messages, which will extend the client expiration times.
     * @param clientJoinMessage - Client join message (from dataContent)
     * @param signalMessage - Ticketed join signal message
     */
    addSequencedSignalClient(clientJoinMessage, signalMessage) {
        const sequencedSignalClient = {
            client: clientJoinMessage.detail,
            referenceSequenceNumber: signalMessage.message.operation
                .referenceSequenceNumber,
            clientConnectionNumber: signalMessage.message.operation.clientConnectionNumber,
            exp: Date.now() + this.serviceConfiguration.deli.clientTimeout,
        };
        this.sequencedSignalClients.set(clientJoinMessage.clientId, sequencedSignalClient);
    }
}
exports.DeliLambda = DeliLambda;
//# sourceMappingURL=lambda.js.map