"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MoiraLambda = void 0;
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const sha_js_1 = __importDefault(require("sha.js"));
const axios_1 = __importDefault(require("axios"));
class MoiraLambda {
    constructor(context, serviceConfiguration, tenantId, documentId) {
        this.context = context;
        this.serviceConfiguration = serviceConfiguration;
        this.tenantId = tenantId;
        this.documentId = documentId;
        this.pending = new Map();
        this.current = new Map();
    }
    handler(message) {
        const boxcar = (0, server_services_core_1.extractBoxcar)(message);
        for (const baseMessage of boxcar.contents) {
            if (baseMessage.type === server_services_core_1.SequencedOperationType) {
                const value = baseMessage;
                // Remove traces and serialize content before writing to mongo.
                value.operation.traces = [];
                const topic = `${value.tenantId}/${value.documentId}`;
                let pendingMessages = this.pending.get(topic);
                if (!pendingMessages) {
                    pendingMessages = [];
                    this.pending.set(topic, pendingMessages);
                }
                pendingMessages.push(value);
            }
        }
        this.pendingOffset = message;
        this.sendPending();
        return undefined;
    }
    close() {
        this.pending.clear();
        this.current.clear();
    }
    sendPending() {
        // If there is work currently being sent or we have no pending work return early
        if (this.current.size > 0 || this.pending.size === 0) {
            return;
        }
        // Swap current and pending
        const temp = this.current;
        this.current = this.pending;
        this.pending = temp;
        const batchOffset = this.pendingOffset;
        const allProcessed = [];
        // Process all the batches + checkpoint
        for (const [, messages] of this.current) {
            const processP = this.processMoiraCoreParallel(messages);
            allProcessed.push(processP);
        }
        Promise.all(allProcessed)
            .then(() => {
            this.current.clear();
            this.context.checkpoint(batchOffset);
            this.sendPending();
        })
            .catch((error) => {
            this.context.error(error, { restart: true });
        });
    }
    createDerivedGuid(referenceGuid, identifier) {
        const hexHash = (0, sha_js_1.default)("sha1").update(`${referenceGuid}:${identifier}`).digest("hex");
        return (`${hexHash.substr(0, 8)}-${hexHash.substr(8, 4)}-` +
            `${hexHash.substr(12, 4)}-${hexHash.substr(16, 4)}-${hexHash.substr(20, 12)}`);
    }
    async processMoiraCoreParallel(messages) {
        var _a, _b, _c, _d;
        const processedMessages = new Map();
        for (const message of messages) {
            if (((_a = message === null || message === void 0 ? void 0 : message.operation) === null || _a === void 0 ? void 0 : _a.type) === "op") {
                const contents = JSON.parse(message.operation.contents);
                const opData = (_d = (_c = (_b = contents.contents) === null || _b === void 0 ? void 0 : _b.contents) === null || _c === void 0 ? void 0 : _c.content) === null || _d === void 0 ? void 0 : _d.contents;
                if (opData && opData.op === 0 && opData.changeSet !== undefined) {
                    // At this point is checked to be submitted to Moira
                    const branchGuid = contents.contents.contents.content.address;
                    const currentProcessing = processedMessages.get(branchGuid);
                    if (currentProcessing) {
                        processedMessages.set(branchGuid, currentProcessing.then(async () => this.processMoiraCore(branchGuid, opData, message)));
                    }
                    else {
                        processedMessages.set(branchGuid, this.processMoiraCore(branchGuid, opData, message));
                    }
                }
            }
        }
        return Promise.all(processedMessages.values());
    }
    async processMoiraCore(branchGuid, opData, message) {
        var _a;
        const commitGuid = opData.guid;
        const logMessage = `MH Commit: branch: ${branchGuid},
        commit ${commitGuid},
        changeSet:  ${JSON.stringify(opData.changeSet, undefined, 2)}`;
        (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.info(logMessage);
        if (this.serviceConfiguration.enableLumberjack) {
            server_services_telemetry_1.Lumberjack.info(logMessage, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId));
        }
        let parentCommitGuid = opData.referenceGuid;
        // Create a branch for the first commit that does not yet reference any other commit
        if (opData.referenceGuid === "") {
            parentCommitGuid = await this.createBranch(branchGuid);
        }
        await this.createCommit(commitGuid, parentCommitGuid, branchGuid, opData, message);
    }
    async createBranch(branchGuid) {
        var _a, _b;
        const rootCommitGuid = this.createDerivedGuid(branchGuid, "root");
        const branchCreationResponse = await axios_1.default.post(`${this.serviceConfiguration.moira.endpoint}/branch`, {
            guid: branchGuid,
            rootCommitGuid,
            meta: {},
            created: 0,
        });
        const lumberProperties = Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { [server_services_telemetry_1.CommonProperties.statusCode]: branchCreationResponse.status });
        if (branchCreationResponse.status === 200) {
            const logMessage = `Branch with guid: ${branchGuid} created`;
            (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.info(logMessage);
            if (this.serviceConfiguration.enableLumberjack) {
                server_services_telemetry_1.Lumberjack.info(logMessage, lumberProperties);
            }
        }
        else {
            const logMessage = `Branch with guid ${branchGuid} failed`;
            (_b = this.context.log) === null || _b === void 0 ? void 0 : _b.error(logMessage);
            if (this.serviceConfiguration.enableLumberjack) {
                server_services_telemetry_1.Lumberjack.error(logMessage, lumberProperties);
            }
        }
        return rootCommitGuid;
    }
    async createCommit(commitGuid, parentGuid, branchGuid, opData, message) {
        var _a, _b, _c;
        try {
            const commitData = {
                guid: commitGuid,
                branchGuid,
                parentGuid,
                meta: {
                    remoteHeadGuid: opData.remoteHeadGuid,
                    localBranchStart: opData.localBranchStart,
                    sequenceNumber: message.operation.sequenceNumber,
                    minimumSequenceNumber: message.operation.minimumSequenceNumber,
                },
            };
            const commitCreationResponse = await axios_1.default.post(`${this.serviceConfiguration.moira.endpoint}/branch/${branchGuid}/commit`, Object.assign(Object.assign({}, commitData), { changeSet: JSON.stringify(opData.changeSet), rebase: true }));
            const lumberProperties = Object.assign(Object.assign({}, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId)), { [server_services_telemetry_1.CommonProperties.statusCode]: commitCreationResponse.status });
            if (commitCreationResponse.status === 200) {
                const logMessage = `Commit created ${JSON.stringify(commitData)}`;
                (_a = this.context.log) === null || _a === void 0 ? void 0 : _a.info(logMessage);
                if (this.serviceConfiguration.enableLumberjack) {
                    server_services_telemetry_1.Lumberjack.info(logMessage, lumberProperties);
                }
            }
            else {
                const logMessage = `Commit failed ${JSON.stringify(commitData)}`;
                (_b = this.context.log) === null || _b === void 0 ? void 0 : _b.error(logMessage);
                if (this.serviceConfiguration.enableLumberjack) {
                    server_services_telemetry_1.Lumberjack.error(logMessage, lumberProperties);
                }
            }
        }
        catch (e) {
            const logMessage = `Commit failed. ${e.message}`;
            (_c = this.context.log) === null || _c === void 0 ? void 0 : _c.error(logMessage);
            if (this.serviceConfiguration.enableLumberjack) {
                server_services_telemetry_1.Lumberjack.error(logMessage, (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId), e);
            }
        }
    }
}
exports.MoiraLambda = MoiraLambda;
//# sourceMappingURL=lambda.js.map