"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.SummaryReader = void 0;
const common_utils_1 = require("@fluidframework/common-utils");
const server_services_client_1 = require("@fluidframework/server-services-client");
const server_services_core_1 = require("@fluidframework/server-services-core");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
/**
 * Git specific implementation of ISummaryReader
 */
class SummaryReader {
    constructor(tenantId, documentId, summaryStorage, enableWholeSummaryUpload, maxRetriesOnError = 6) {
        this.tenantId = tenantId;
        this.documentId = documentId;
        this.summaryStorage = summaryStorage;
        this.enableWholeSummaryUpload = enableWholeSummaryUpload;
        this.maxRetriesOnError = maxRetriesOnError;
        this.lumberProperties = (0, server_services_telemetry_1.getLumberBaseProperties)(this.documentId, this.tenantId);
    }
    /**
     * Reads the most recent version of summary for a document. In case the storage is having trouble processing the
     * request, returns a set of defaults with fromSummary flag set to false.
     */
    async readLastSummary() {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
        const summaryReaderMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.SummaryReader);
        summaryReaderMetric.setProperties(this.lumberProperties);
        if (this.enableWholeSummaryUpload) {
            try {
                let wholeFlatSummary;
                try {
                    wholeFlatSummary = await (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getSummary(server_services_client_1.LatestSummaryId), "readWholeSummary_getSummary", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError);
                }
                catch (error) {
                    server_services_telemetry_1.Lumberjack.warning(`Error fetching summary with ID ${server_services_client_1.LatestSummaryId}. Will try with explicit ref.`, this.lumberProperties, error);
                    wholeFlatSummary = undefined;
                }
                if (!wholeFlatSummary) {
                    const existingRef = await (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getRef(encodeURIComponent(this.documentId)), "readWholeSummary_getRef", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError);
                    wholeFlatSummary = await (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getSummary(existingRef.object.sha), "readWholeSummary_getSummary", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError);
                }
                const normalizedSummary = (0, server_services_client_1.convertWholeFlatSummaryToSnapshotTreeAndBlobs)(wholeFlatSummary);
                // Obtain IDs of specific fields from the downloaded summary
                const attributesBlobId = (_b = (_a = normalizedSummary.snapshotTree.trees[".protocol"]) === null || _a === void 0 ? void 0 : _a.blobs) === null || _b === void 0 ? void 0 : _b.attributes;
                const scribeBlobId = (_d = (_c = normalizedSummary.snapshotTree.trees[".serviceProtocol"]) === null || _c === void 0 ? void 0 : _c.blobs) === null || _d === void 0 ? void 0 : _d.scribe;
                const deliBlobId = (_f = (_e = normalizedSummary.snapshotTree.trees[".serviceProtocol"]) === null || _e === void 0 ? void 0 : _e.blobs) === null || _f === void 0 ? void 0 : _f.deli;
                const opsBlobId = (_h = (_g = normalizedSummary.snapshotTree.trees[".logTail"]) === null || _g === void 0 ? void 0 : _g.blobs) === null || _h === void 0 ? void 0 : _h.logTail;
                // Parse specific fields from the downloaded summary
                const attributesContent = attributesBlobId
                    ? normalizedSummary.blobs.get(attributesBlobId)
                    : undefined;
                const scribeContent = scribeBlobId
                    ? normalizedSummary.blobs.get(scribeBlobId)
                    : undefined;
                const deliContent = deliBlobId
                    ? normalizedSummary.blobs.get(deliBlobId)
                    : undefined;
                const opsContent = opsBlobId ? normalizedSummary.blobs.get(opsBlobId) : undefined;
                const attributes = attributesContent
                    ? JSON.parse((0, common_utils_1.bufferToString)(attributesContent, "utf8"))
                    : this.getDefaultAttributes();
                const scribe = scribeContent
                    ? (0, common_utils_1.bufferToString)(scribeContent, "utf8")
                    : this.getDefaultScribe();
                const deli = deliContent
                    ? JSON.parse((0, common_utils_1.bufferToString)(deliContent, "utf8"))
                    : this.getDefaultDeli();
                const messages = opsContent
                    ? JSON.parse((0, common_utils_1.bufferToString)(opsContent, "utf8"))
                    : this.getDefaultMesages();
                summaryReaderMetric.setProperties({
                    [server_services_telemetry_1.CommonProperties.minLogtailSequenceNumber]: Math.min(...messages.map((message) => message.sequenceNumber)),
                    [server_services_telemetry_1.CommonProperties.maxLogtailSequenceNumber]: Math.max(...messages.map((message) => message.sequenceNumber)),
                    [server_services_telemetry_1.CommonProperties.lastSummarySequenceNumber]: deli.sequenceNumber,
                    [server_services_telemetry_1.CommonProperties.clientCount]: (_j = deli.clients) === null || _j === void 0 ? void 0 : _j.length,
                });
                summaryReaderMetric.success(`Successfully read whole summary`);
                return {
                    protocolHead: attributes.sequenceNumber,
                    scribe,
                    messages,
                    fromSummary: true,
                };
            }
            catch (error) {
                summaryReaderMetric.error(`Returning default summary due to error when reading whole summary`, error);
                return this.getDefaultSummaryState();
            }
        }
        else {
            try {
                const existingRef = await (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getRef(encodeURIComponent(this.documentId)), "readSummary_getRef", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError);
                const [attributesContent, scribeContent, deliContent, opsContent] = await Promise.all([
                    (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getContent(existingRef.object.sha, ".protocol/attributes"), "readSummary_getProtocolAttributesContent", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError).catch(() => undefined),
                    (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getContent(existingRef.object.sha, ".serviceProtocol/scribe"), "readSummary_getServiceProtocolScribeContent", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError).catch(() => undefined),
                    (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getContent(existingRef.object.sha, ".serviceProtocol/deli"), "readSummary_getServiceProtocolDeliContent", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError).catch(() => undefined),
                    (0, server_services_core_1.requestWithRetry)(async () => this.summaryStorage.getContent(existingRef.object.sha, ".logTail/logTail"), "readSummary_getLogTailContent", this.lumberProperties, server_services_core_1.shouldRetryNetworkError, this.maxRetriesOnError).catch(() => undefined),
                ]);
                const attributes = attributesContent
                    ? JSON.parse((0, common_utils_1.toUtf8)(attributesContent.content, attributesContent.encoding))
                    : this.getDefaultAttributes();
                const scribe = scribeContent
                    ? (0, common_utils_1.toUtf8)(scribeContent.content, scribeContent.encoding)
                    : this.getDefaultScribe();
                const deli = deliContent
                    ? JSON.parse((0, common_utils_1.toUtf8)(deliContent.content, deliContent.encoding))
                    : this.getDefaultDeli();
                const messages = opsContent
                    ? JSON.parse((0, common_utils_1.toUtf8)(opsContent.content, opsContent.encoding))
                    : this.getDefaultMesages();
                summaryReaderMetric.setProperties({
                    [server_services_telemetry_1.CommonProperties.minLogtailSequenceNumber]: Math.min(...messages.map((message) => message.sequenceNumber)),
                    [server_services_telemetry_1.CommonProperties.maxLogtailSequenceNumber]: Math.max(...messages.map((message) => message.sequenceNumber)),
                    [server_services_telemetry_1.CommonProperties.lastSummarySequenceNumber]: deli.sequenceNumber,
                    [server_services_telemetry_1.CommonProperties.clientCount]: (_k = deli.clients) === null || _k === void 0 ? void 0 : _k.length,
                });
                summaryReaderMetric.success(`Successfully read summary`);
                return {
                    protocolHead: attributes.sequenceNumber,
                    scribe,
                    messages,
                    fromSummary: true,
                };
            }
            catch (error) {
                summaryReaderMetric.error(`Returning default summary due to error when reading summary`, error);
                return this.getDefaultSummaryState();
            }
        }
    }
    getDefaultSummaryState() {
        return {
            protocolHead: 0,
            scribe: "",
            messages: [],
            fromSummary: false,
        };
    }
    getDefaultAttributes() {
        server_services_telemetry_1.Lumberjack.info("Using default attributes when reading summary.", this.lumberProperties);
        return {
            sequenceNumber: 0,
            minimumSequenceNumber: 0,
            // "term" was an experimental feature that is being removed.  The only safe value to use is 1.
            term: 1,
        };
    }
    getDefaultScribe() {
        server_services_telemetry_1.Lumberjack.info("Using default scribe when reading summary.", this.lumberProperties);
        return "";
    }
    getDefaultDeli() {
        server_services_telemetry_1.Lumberjack.info("Using default deli state when reading summary.", this.lumberProperties);
        return {
            clients: undefined,
            durableSequenceNumber: 0,
            logOffset: 0,
            sequenceNumber: 0,
            signalClientConnectionNumber: 0,
            expHash1: "",
            lastSentMSN: undefined,
            nackMessages: undefined,
            successfullyStartedLambdas: [],
            checkpointTimestamp: undefined,
        };
    }
    getDefaultMesages() {
        server_services_telemetry_1.Lumberjack.info("Using default messages when reading summary.", this.lumberProperties);
        return [];
    }
}
exports.SummaryReader = SummaryReader;
//# sourceMappingURL=summaryReader.js.map